"""Functions for reading and writing POSCAR files.

Also defines the POSCARError specific exception, as well as some subclasses.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-19'
__license__ = 'GPLv3+'

from collections import defaultdict
from contextlib import AbstractContextManager
from io import TextIOBase
import logging
from pathlib import Path

import numpy as np

from viperleed.calc.classes import atom as calc_atom
from viperleed.calc.classes import slab as calc_slab

_LOGGER = logging.getLogger(__name__)


class POSCARError(Exception):
    """A generic exception related to a POSCAR file."""


class POSCARSyntaxError(POSCARError):
    """An exceptions for syntax errors in POSCAR files."""


def read(filename='POSCAR'):
    """Return a Slab with the contents of a POSCAR file.

    Parameters
    ----------
    filename : str or Path or TextIOBase, optional
        The file to read from. May be an (open) file object.
        The default is 'POSCAR'.

    Returns
    -------
    slab : Slab
        The slab object. Notice that atoms are normally sorted the              # TODO: discuss if OK
        same way as they were in filename. The only exception is when
        the file contains atoms that are non-element-contiguous, i.e.,
        there are multiple "blocks" with the same chemical element. In
        the latter case, atoms are re-sorted to be element-contiguous,
        in order of appearance of the chemical elements.

    Raises
    ------
    FileNotFoundError
        If filename is not the path to a valid file.
    POSCARSyntaxError
        If unit cell vectors have invalid shape (not enough
        vectors or not enough components).
    POSCARSyntaxError
        If the file does not contain the (normally optional)
        line of chemical-element labels, or if this line is
        inconsistent with the one specifying the atom counts.
    POSCARSyntaxError
        If no (or too few) atomic coordinate is read.
    """
    if isinstance(filename, TextIOBase):
        poscar = POSCARStreamReader(filename)
        return poscar.read()

    with POSCARFileReader(filename) as poscar:
        return poscar.read()


def write(slab, filename='CONTCAR', comments='none', silent=False):
    """Write a POSCAR-style file from a slab.

    If a file named 'POSCAR' exists in the current folder, its first,
    i.e., comment line is copied over.

    Parameters
    ----------
    slab : Slab
        The Slab object to write.
    filename : str or Path or TextIOBase, optional
        The file(name) to write to. The default is 'CONTCAR'.
    comments : str, optional
        Defines whether additional information for each atom
        should be printed:
        - 'none' writes a normal POSCAR (Default)
        - 'all' all annotations
        - 'nodir' all annotations except directions relating
          to a or b (meant to be used with POSCAR_oricell)
        - 'bulk' is like 'none' but writes the space group
          (meant to be used with POSCAR_bulk).
    silent : bool, optional
        If True, will print less to log.

    Raises
    -------
    OSError
        If writing to filename fails.
    """
    slab.sort_by_element()   # This is the minimum that has to happen

    if isinstance(filename, TextIOBase):
        poscar = POSCARStreamWriter(filename, comments=comments)
        poscar.write(slab)
        return

    try:  # pylint: disable=too-many-try-statements
        with POSCARFileWriter(filename, comments=comments) as poscar:
            poscar.write(slab)
    except OSError:
        _LOGGER.error(f'Failed to write {filename}')
        raise
    if not silent:
        _LOGGER.debug(f'Wrote to {filename} successfully')


class POSCARReader:
    """Base class for reading POSCAR structures into a Slab."""

    def __init__(self, source, *, ucell_eps=1e-4):
        """Initialize instance.

        Parameters
        ----------
        source : object
            Positional-only. The source from which the POSCAR
            structure should be read.
        ucell_eps : float, optional
            Cartesian tolerance (Angstrom). All the unit cell
            Cartesian components smaller than ucell_eps will
            be zeroed exactly. Default is 1e-4.
        """
        self._source = source
        self.ucell_eps = ucell_eps

    @property
    def stream(self):
        """Return the stream from which to read structure information."""
        raise NotImplementedError

    def read(self):
        """Return a slab with info read from source."""
        slab = calc_slab.Slab()

        next(self.stream, None)  # Skip first line: comment
        slab.poscar_scaling = float(next(self.stream, '1').split()[0])          # TODO: POSCAR wiki says it's more complex!

        self._read_unit_cell(slab)    # Three lines of unit cell
        element_info = self._read_elements(slab)
        cartesian, _ = self._read_cartesian_and_group(slab)
        positions = self._read_atom_coordinates(slab, cartesian)
        self._make_atoms(slab, positions, *element_info)
        return slab

    @staticmethod
    def _make_atoms(slab, positions, elements, element_counts):
        """Populate slab with atoms from their fractional coordinates.

        Parameters
        ----------
        slab : Slab
            The slab to which atoms are to be added.
        positions : Sequence
            Each element is the (a, b, c) fractional coordinate of
            an atom. There should be exactly sum(element_counts)
            coordinates. See also element_counts.
        elements : Sequence
            Each element is a string, corresponding to the distinct
            chemical species to be added. Chemical species need not
            be unique.
        element_counts : Sequence
            Same number of items as elements. element_counts[i] is
            the number of atoms with element elements[i].

        Raises
        -------
        ValueError
            If the number of atomic coordinates given is not equal
            to the sum of element_counts, if any position is not
            a three-element Sequence, and if element_counts and
            elements have different number of items.
        """
        n_positions, n_coords = np.shape(positions)
        if n_positions != sum(element_counts):
            raise ValueError('Inconsistent number of atomic fractional '
                             f'coordinates ({n_positions}) and chemical '
                             f'element counts ({sum(element_counts)}).')
        if n_coords != 3:
            raise ValueError('Invalid fractional coordinates. Expected '
                             f'3 components, found {n_coords}.')
        if len(elements) != len(element_counts):
            raise ValueError('Inconsistent number of items in elements '
                             f'({len(elements)} and in element_counts '
                             f'({len(element_counts)}')

        positions = iter(positions)
        atoms = defaultdict(list)  # Merge same-element blocks
        for element, n_atoms in zip(elements, element_counts):
            for n_added, fractional_pos in enumerate(positions, start=1):
                atoms[element].append(
                    calc_atom.Atom(element, fractional_pos, 0, slab)
                    )
                if n_added >= n_atoms:
                    break
        slab.atlist.clear()
        slab.atlist.strict = False  # All atoms have the same .num
        slab.atlist.extend(at for el_ats in atoms.values() for at in el_ats)
        slab.update_atom_numbers()  # Also updates atlist map
        slab.atlist.strict = True
        slab.update_element_count()
        slab.update_cartesian_from_fractional()

    def _read_atom_coordinates(self, slab, cartesian):
        """Return an array of fractional atomic coordinates read from file_."""
        n_atoms = sum(slab.n_per_elem.values())
        positions = []
        for line in self.stream:
            coordinates = line.split()
            if not coordinates:
                _LOGGER.debug('POSCAR: Empty line found; '
                              'stopping position readout')
                break
            try:
                positions.append([float(c) for c in coordinates[:3]])
            except ValueError:  # Reached the optional "Lattice velocities"
                _LOGGER.debug(f'POSCAR: no coordinates in {line!r}; '
                              'stopping position readout')
                break
            if len(positions) >= n_atoms:
                break

        if not positions:
            raise POSCARSyntaxError('POSCAR: No atomic coordinates found.')
        if len(positions) < n_atoms:
            raise POSCARSyntaxError(
                'POSCAR: Too few atomic coordinates. Found '
                f'{len(positions)}, expected {n_atoms}'
                )

        # pylint: disable=redefined-variable-type
        # Disabled because used only locally, and clear enough.
        positions = np.array(positions)
        if cartesian:
            uc_inv = np.linalg.inv(slab.ucell.T)
            positions = positions.dot(uc_inv)
        positions[:, :2] %= 1.0  # Collapse coordinates in a and b
        return positions

    def _read_elements(self, slab):
        """Return chemical elements and counts, and update slab."""
        # Line of element labels, then their counts
        elements = [el.capitalize() for el in next(self.stream, '').split()]

        # Element symbols may be terminated by a '/' + some hash; remove it
        # See https://www.vasp.at/forum/viewtopic.php?t=19113
        elements = [el.split('/')[0] for el in elements]

        # Element labels line is optional, but we need it
        try:
            _ = [int(e) for e in elements]
        except ValueError:  # Some non-numeric text. OK
            pass
        else:  # Only numbers. Probably just the element counts
            elements = []
        if not elements:
            raise POSCARSyntaxError(
                'POSCAR: Element labels line not found. This is an optional '
                'line in the POSCAR specification, but ViPErLEED needs it.'
                )
        if any(not el for el in elements):
            raise POSCARSyntaxError(
                'POSCAR: Empty element label found. This is not allowed.'
                )

        element_counts = [int(c) for c in next(self.stream, '').split()]
        if len(element_counts) != len(elements):
            raise POSCARSyntaxError('Length of element list does not match '
                                    'length of atoms-per-element list')
        n_per_elem = defaultdict(int)
        for element, n_atoms in zip(elements, element_counts):
            n_per_elem[element] += n_atoms
        slab.n_per_elem = dict(n_per_elem)
        if len(slab.n_per_elem) != len(elements):
            logging.warning(
                'POSCAR: atoms are not element-contiguous, i.e., '
                'there are multiple blocks with the same chemical-'
                f'species label ({elements}). This structure will '
                'not be preserved, i.e., atoms will be re-sorted '
                f'(to {list(slab.n_per_elem.keys())}) so that only '
                'one block for each element is present.'
                )
        return elements, element_counts

    def _read_cartesian_and_group(self, slab):
        """Read the line containing the "Cartesian"/"Direct" information.

        slab : Slab
            The slab object in which information is to be stored.
            The only attribute modified is .preprocessed, depending
            on whether there is suitable plane-group information.

        Returns
        -------
        cartesian : bool
            Whether the atomic coordinates in the POSCAR file are
            expressed as fractional (False) or Cartesian (True).
        plane_group : str
            The plane group found in the POSCAR file. It will be
            'unknown' if no plane group information exists, or in
            case the information read is not a valid plane group.
        """
        # Skip the optional "S[elective Dynamics]" line
        cartesian_line = next(self.stream, '').lower()
        if cartesian_line.startswith('s'):
            _LOGGER.debug("POSCAR: skipping 'Selective dynamics' line (no. 9)")
            cartesian_line = next(self.stream, '').lower()
        cartesian = cartesian_line.startswith(('c', 'k'))

        # Check whether POSCAR was pre-processed, i.e., whether
        # the 'Plane group = ...' comment is already there
        plane_group = 'unknown'
        if all(s in cartesian_line for s in ('plane group = ', '  n')):
            _, plane_group_str = cartesian_line.split('plane group = ')
            plane_group_str = plane_group_str.split('  n')[0]
            plane_group_str = plane_group_str.split('(')[0].strip()
            if not plane_group_str.startswith('*'):
                slab.preprocessed = True
            plane_group = plane_group_str
        return cartesian, plane_group

    def _read_unit_cell(self, slab):
        """Read unit-cell vectors into slab from the next three lines."""
        ucell = []
        for line in self.stream:
            try:
                ucell.append([float(coord) for coord in line.split()])
            except ValueError:
                break
            if len(ucell) == 3:
                break
        slab.ucell = np.array(ucell).T * slab.poscar_scaling
        if slab.ucell.shape != (3, 3):
            try:
                n_vec, n_comp = slab.ucell.shape
            except ValueError:
                _err = ('Invalid unit-cell vectors. Expected shape (3, 3), '
                        f'found {slab.ucell.shape} instead')
            else:
                _err = ('Invalid unit-cell vectors: not enough vectors '
                        f'({n_vec}/3) or not enough Cartesian components '
                        f'({n_comp}/3).')
            _LOGGER.error(_err)
            raise POSCARSyntaxError(_err)

        slab.ucell_ori = slab.ucell.copy()
        slab.ucell[abs(slab.ucell) < self.ucell_eps] = 0.                       # TODO: was done only in x,y. OK to do all?


class POSCARFileReader(AbstractContextManager, POSCARReader):
    """A context manager for reading POSCAR files into a slab."""

    def __init__(self, filename, *, ucell_eps=1e-4):
        """Initialize instance.

        Parameters
        ----------
        filename : str or Path
            The path to the file to be read.
        ucell_eps : float, optional
            Cartesian tolerance (Angstrom). All the unit cell
            Cartesian components smaller than ucell_eps will
            be zeroed exactly. Default is 1e-4.
        """
        super().__init__(Path(filename), ucell_eps=ucell_eps)
        self._file_object = None

    @property
    def filename(self):
        """Return the name of the file from which to read."""
        return self._source  # From POSCARReader

    @property
    def stream(self):
        """Return the stream from which to read structure information."""
        return self._file_object

    def __enter__(self):
        """Enter context."""
        if not self.filename.is_file():
            _LOGGER.error('POSCAR not found.')
            raise FileNotFoundError(
                f'POSCAR file at {self.filename} not found.'
                )
        self._file_object = self.filename.open('r', encoding='utf-8')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Close file object when exiting."""
        try:
            self._file_object.close()
        except AttributeError:
            pass
        return super().__exit__(exc_type, exc_val, exc_tb)

    def read(self):
        """Return a slab with info read from file."""
        if self._file_object is None or self._file_object.closed:
            raise ValueError(
                'I/O operation on closed file. Use\n'
                f'with {type(self).__name__}(filename) as poscar:\n'
                '    slab = poscar.read()'
                )
        return super().read()


class POSCARStreamReader(POSCARReader):
    """Class for reading a POSCAR structure from an open stream."""

    @property
    def stream(self):
        """Return the stream from which to read structure information."""
        return self._source

    def read(self):
        """Return a slab with info read from file."""
        if self.stream.closed:
            raise ValueError('I/O operation on closed stream.')
        return super().read()


class POSCARWriter:
    """Base class for writing a Slab to POSCAR format."""

    def __init__(self, target, *, comments='none', **__kwargs):
        """Initialize instance.

        Parameters
        ----------
        target : object
            Positional-only. An object to be used for writing the
            structure in POSCAR format. Subclasses are more specific
            as to which objects they require.
        comments : str, optional
            Defines whether additional information for each atom
            should be printed:
            - 'none' writes a normal POSCAR (Default)
            - 'all' all annotations
            - 'nodir' all annotations except directions relating
              to a or b (meant to be used with POSCAR_oricell)
            - 'bulk' is like 'none' but writes the space group
              (meant to be used with POSCAR_bulk).
        **__kwargs : object, optional
            Other unused keyword-only arguments.

        Returns
        -------
        None.
        """
        self._target = target
        self.comments = comments
        self.slab = None
        self.header = '\n'  # Default is no header

    @property
    def stream(self):
        """Return the stream to which the POSCAR should be written."""
        raise NotImplementedError

    def write(self, slab):
        """Write the info in slab to the self.stream as a POSCAR."""
        self.stream.write(self.header)
        self.stream.writelines(self._get_scaling(slab))
        self.stream.writelines(self._get_unit_cell(slab))
        self.stream.writelines(self._get_elements_and_numbers(slab))
        self._write_selective_dynamics_line()
        self.stream.writelines(self._get_cartesian_and_group_line(slab))
        self.stream.writelines(self._get_atom_coordinates(slab))

    @staticmethod
    def _get_scaling(slab):
        """Yield the 'poscar-scaling' line."""
        yield f'{slab.poscar_scaling:10.6f}\n'

    @staticmethod
    def _get_unit_cell(slab):
        """Yield one line per unit cell vector."""
        for vector in slab.ucell.T:
            vec = ''.join(f'{v / slab.poscar_scaling:22.16f}' for v in vector)
            yield vec + '\n'

    @staticmethod
    def _get_elements_and_numbers(slab):
        """Yield the two lines of chemical species and their counts."""
        yield ''.join(f'{el:>5}' for el in slab.elements) + '\n'
        yield ''.join(f'{n:>5}' for n in slab.n_per_elem.values()) + '\n'

    def _get_atom_coordinates(self, slab):
        """Yield a line of coordinates and other info for each atom in slab."""
        more_than_one = (link for link in slab.linklists if len(link) > 1)
        linklists = {id(link): i+1 for i, link in enumerate(more_than_one)}
        for i, atom in enumerate(slab):
            line = ''.join(f'{coord:20.16f}' for coord in atom.pos)
            line += self._get_comments_for_atom(atom, i, linklists)
            yield line + '\n'

    def _get_cartesian_and_group_line(self, slab):
        """Yield the line for "Direct", group and other column headers."""
        if self.comments == 'none':
            yield 'Direct\n'
            return

        line = f'{"Direct":<16}Plane group = '
        if self.comments == 'nodir':
            line += '*'

        group = slab.planegroup
        if (slab.planegroup in ["pm", "pg", "cm", "rcm", "pmg"]
                and self.comments != 'nodir'):
            line += f"{group}{slab.orisymplane.par}"
        else:
            line += group

        if slab.planegroup != slab.foundplanegroup.split('[')[0]:
            line += ' (found '
            if '[' in slab.foundplanegroup and self.comments == 'nodir':
                line += slab.foundplanegroup.split('[')[0]
            else:
                line += slab.foundplanegroup
            line += ')'
        line = line.ljust(60)
        if self.comments == 'all':
            line += (f'{"N":>5}{"SiteLabel":>12}{"Layer":>7}'
                     f'{"Linking":>9}{"FreeDir":>12}')
        elif self.comments == 'nodir':
            line += (f'{"N":>5}{"SiteLabel":>12}{"Layer":>7}'
                     f'{"Linking":>9}{"FreeDir":>12}: see POSCAR')
        yield line + '\n'

    def _get_comments_for_atom(self, atom, atom_index, linklists):
        """Return a line of comments for atom at atom_index."""
        if self.comments in ('none', 'bulk'):
            # Only coordinates
            return ''

        line = (f'{atom_index+1:>5}'                  # N
                # f'{atom.el}{NperElCount}'.rjust(9)  # NperEl
                f'{atom.site.label:>12}'              # SiteLabel
                f'{atom.layer.num + 1:>7}')           # Layer
        if len(atom.linklist) <= 1:                   # Linking
            linked = 'none'
        else:
            linked = linklists[id(atom.linklist)]
        line += f'{linked:>9}'
        if self.comments == 'nodir':
            return line
        #                                              # FreeDir
        _free_dir = ""
        if atom.is_bulk:
            _free_dir = 'bulk'
        elif isinstance(atom.freedir, np.ndarray):
            _free_dir = str(atom.freedir)  # has to be made into a string
        elif atom.freedir == 0:
            _free_dir = 'locked'
        else:
            _free_dir = 'free'
        line += f"{_free_dir:>12}"
        return line

    def _write_selective_dynamics_line(self):
        """Write the 'Selective dynamics' line needed for DFT relaxation."""


class POSCARFileWriter(AbstractContextManager, POSCARWriter):
    """A context manager for writing a slab to POSCAR."""

    def __init__(self, filename, *, comments='none'):
        """Initialize instance.

        Parameters
        ----------
        filename : str or Path
            The path to the file to be read.
        comments : str, optional
            Defines whether additional information for each atom
            should be printed:
            - 'none' writes a normal POSCAR (Default)
            - 'all' all annotations
            - 'nodir' all annotations except directions relating
              to a or b (meant to be used with POSCAR_oricell)
            - 'bulk' is like 'none' but writes the space group
              (meant to be used with POSCAR_bulk).

        Returns
        -------
        None.
        """
        super().__init__(Path(filename), comments=comments)
        self._file_object = None

        # Take header line from existing POSCAR, or use a dummy header
        poscar = Path('POSCAR')
        if poscar.is_file():
            with poscar.open('r', encoding='utf-8') as _file:
                self.header = _file.readline()
        else:
            self.header = 'unknown'
        if not self.header.endswith('\n'):
            self.header += '\n'

    @property
    def filename(self):
        """Return the path to the file to be written."""
        return self._target

    @property
    def stream(self):
        """Return the stream to which the POSCAR should be written."""
        return self._file_object

    def __enter__(self):
        """Enter context."""
        self._file_object = self.filename.open('w', encoding='utf-8')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Close file object when exiting."""
        try:
            self._file_object.close()
        except AttributeError:
            pass
        super().__exit__(exc_type, exc_val, exc_tb)

    def write(self, slab):
        """Write the info contained in slab to file."""
        if self._file_object is None or self._file_object.closed:
            raise ValueError(
                'I/O operation on closed file. Use\n'
                f'with {type(self).__name__}(filename) as poscar:\n'
                '    poscar.write(slab)'
                )
        super().write(slab)


class POSCARStreamWriter(POSCARWriter):
    """Class for writing a Slab to an open stream in POSCAR format."""

    @property
    def stream(self):
        """Return the stream to which the POSCAR should be written."""
        return self._target

    def write(self, slab):
        """Write the info contained in slab to the current stream."""
        if self.stream.closed:
            raise ValueError('I/O operation on closed stream')
        super().write(slab)


class VASPPOSCARWriter(POSCARStreamWriter):
    """Class for writing a Slab to a POSCAR, ready for VASP relaxation."""

    def __init__(self, stream, *, comments='none', relax_info=None):
        """Initialize instance.

        Parameters
        ----------
        stream : TextIOBase
            The stream to which the POSCAR will be written.
        comments : str, optional
            Defines whether additional information for each atom
            should be printed:
            - 'none' writes a normal POSCAR (Default)
            - 'all' all annotations
            - 'nodir' all annotations except directions relating
              to a or b (meant to be used with POSCAR_oricell)
            - 'bulk' is like 'none' but writes the space group
              (meant to be used with POSCAR_bulk).
            This argument is ignored if relax_info is given.
        relax_info : dict or None, optional
            Contains optional information on how the slab should
            be relaxed. Keys in use:
            - 'above_c' : float, optional
                Only atoms above this c fraction are allowed to move.
                Should be between zero and one (both included). If not
                given, all atoms can move.
            - 'c_only' : bool, optional
                Are movements allowed only along c (True) or also in
                plane (False)? Default is True.

        Raises
        ------
        ValueError
            If the 'above_c' key of relax_info is not a
            floating-point number between 0 and 1 (included)
        """
        super().__init__(stream, comments=comments)
        if relax_info is None:
            relax_info = {}
        self.relax_info = relax_info
        self._check_relax_info()

    def _check_relax_info(self):
        """Raise if self.relax_info contains inappropriate values."""
        above_c = self.relax_info.get('above_c', 0)
        if not 0 <= above_c <= 1:
            raise ValueError('Invalid relax_info: "above_c" '
                             'must be in range [0, 1].')

    def _write_selective_dynamics_line(self):
        """Write the 'Selective dynamics' line needed for DFT relaxation."""
        if self.relax_info:
            self.stream.write('Selective dynamics\n')

    def _get_comments_for_atom(self, atom, atom_index, linklists):
        """Return a line of comments for atom at atom_index."""
        if not self.relax_info:
            return super()._get_comments_for_atom(atom, atom_index, linklists)

        c_cutoff = self.relax_info.get('above_c', -np.inf)
        relax_only_along_c = self.relax_info.get('c_only', True)

        if atom.pos[2] <= c_cutoff:
            return '   F   F   F'
        if relax_only_along_c:
            return '   F   F   T'
        return '   T   T   T'



"""Module cli of viperleed.utilities.

Defines the command-line interface for the ViPErLEED utilities.
"""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-08-02'
__license__ = 'GPLv3+'


from viperleed.cli_base import ViPErLEEDCLIWithAutoChildren


class ViPErLEEDUtilitiesCLI(ViPErLEEDCLIWithAutoChildren,
                            cli_name='utilities',
                            help_='run ViPErLEED utilities'):
    """The main command-line interface for the ViPErLEED utilities."""

    def __init__(self):
        """Initialize CLI with some aliases."""
        super().__init__()
        self.add_child_aliases('rearrange-phaseshifts',
                               'rearrange-phaseshift',
                               'rearrange_phaseshifts',
                               'rearrange_phaseshift')



"""ViPErLEED utility: Reorder elements."""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-08-03'
__license__ = 'GPLv3+'

from viperleed.calc.lib import periodic_table
from viperleed.utilities.poscar.base import _PoscarStreamCLI


class ReorderElementsCLI(_PoscarStreamCLI, cli_name='reorder_elements'):
    """Change the sort order of the elements in a POSCAR."""

    long_name = 'reorder elements'

    def add_parser_arguments(self, parser):
        """Add ordering flags as arguments to parser."""
        super().add_parser_arguments(parser)
        order = parser.add_mutually_exclusive_group()
        order.add_argument(
            '--ascending',
            help=('sort elements by atomic number in ascending order. '
                  'This is the default if no argument is given.'),
            action='store_const',
            dest='ordering',
            const=order_ascending,
            )
        order.add_argument(
            '-d', '--descending',
            help='sort elements by atomic number in descending order',
            action='store_const',
            dest='ordering',
            const=order_descending,
            )
        order.add_argument(
            '-a', '--alphabetical',
            help='sort elements alphabetically',
            action='store_const',
            dest='ordering',
            const=order_alphabetical,
            )
        order.add_argument(
            '-c', '--custom',
            help=('Sort elements by custom order. Provide a comma-separated '
                  'list of elements in the desired order.'),
            type=str,
            )

    def parse_cli_args(self, args):
        """Set a default if no argument is given."""
        parsed_args = super().parse_cli_args(args)
        if parsed_args.ordering is None and parsed_args.custom is None:
            parsed_args.ordering = order_ascending
        return parsed_args

    def process_slab(self, slab, args):
        """Reorder elements in slab according to args and return slab."""
        try:
            ordering = args.ordering or make_custom_ordering(args.custom, slab)
        except KeyError as exc:
            self.parser.error(str(exc))

        ordered_elements = sorted(slab.elements, key=ordering)
        slab.n_per_elem = {el: slab.n_per_elem[el] for el in ordered_elements}
        slab.sort_by_z()
        return slab


def make_custom_ordering(custom_order, slab):
    """Return a callable that sorts items according to custom_order."""
    custom_order_map = {el.strip().capitalize(): i
                        for i, el in enumerate(custom_order.split(','))}
    missing_elements = set(slab.elements) - set(custom_order_map)
    if missing_elements:
        # Prevents KeyError later on when accessing the slab elements
        raise KeyError('Missing elements in custom order: '
                       + ','.join(sorted(missing_elements)))

    def order_custom(element_name):
        return custom_order_map[element_name]

    return order_custom


def order_alphabetical(element_name):
    """Return a key for sorting element_name alphabetically."""
    return element_name.lower()


def order_ascending(element_name):
    """Return a key for sorting element_name by ascending atomic number."""
    return periodic_table.get_atomic_number(element_name)


def order_descending(element_name):
    """Return a key for sorting element_name by descending atomic number."""
    return -order_ascending(element_name)


if __name__ == '__main__':
    ReorderElementsCLI.run_as_script()



"""Module run of viperleed.calc.

Defines the main functionality for running a
viperleed calculation from a set of input files.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2019-11-12'  # Was originally tleedm.py
__license__ = 'GPLv3+'

import logging
import os
from pathlib import Path
import shutil
import time

from viperleed import __version__
from viperleed.calc import LOGGER as logger
from viperleed.calc import LOG_PREFIX
from viperleed.calc.classes import rparams
from viperleed.calc.files import parameters, poscar
from viperleed.calc.files.tenserleed import get_tensorleed_path
from viperleed.calc.lib.base import CustomLogFormatter
from viperleed.calc.sections.cleanup import cleanup
from viperleed.calc.sections.cleanup import prerun_clean
from viperleed.calc.sections.initialization import (
    warn_if_slab_has_atoms_in_multiple_c_cells
    )
from viperleed.calc.sections.run_sections import section_loop



def run_calc(system_name=None,
             console_output=True,
             slab=None,
             preset_params=None,
             source=None):
    """Run a ViPErLEED calculation.

    By default, a PARAMETERS and a POSCAR file are expected, but can be
    replaced by passing the `slab` and/or `present_params` kwargs.

    Parameters
    ----------
    system_name : str, optional
        Used as a comment in some output-file headers. If not
        specified, the name of the parent directory is used
        instead. Default is None.
    console_output : bool, optional
        If False, will not add a logging.StreamHandler. Output
        will only be printed to the log file. Default is True.
    slab : Slab, optional
        Start from a pre-existing slab, instead of reading from
        POSCAR. Default is None.
    preset_params : dict, optional
        Parameters to add to the Rparams object after a PARAMETERS file
        has been read. Keys should be attributes of Rparams. Values in
        `preset_params` will overwrite values read from the PARAMETERS
        file, if present in both. If no PARAMETERS file is read,
        parameters will be read exclusively from `preset_params`.
        Default is None.
    source : pathlike, optional
        Path where the tensor-LEED directory can be found, which
        contains all the TensErLEED source code. If not given
        or None, try taking it from the environment variable
        VIPERLEED_TENSORLEED. As a last resort, use the current
        directory. Default is None.

    Returns
    -------
    int
        0: exit without errors.
        1: clean exit through KeyboardInterrupt
        2: exit due to Exception before entering main loop
        3: exit due to Exception during main loop
    """
    os.umask(0)
    # start logger, write to file:
    timestamp = time.strftime("%y%m%d-%H%M%S", time.localtime())

    log_name = f'{LOG_PREFIX}-{timestamp}.log'
    logger.setLevel(logging.INFO)
    logFormatter = CustomLogFormatter()
    fileHandler = logging.FileHandler(log_name, mode="w")
    fileHandler.setFormatter(logFormatter)
    logger.addHandler(fileHandler)
    if console_output:
        consoleHandler = logging.StreamHandler()
        consoleHandler.setFormatter(logFormatter)
        logger.addHandler(consoleHandler)

    logger.info(f"Starting new log: {log_name}\nTime of execution (UTC): "
                + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))
    logger.info(f"This is ViPErLEED version {__version__}\n")

    tmp_manifest = ["SUPP", "OUT", log_name]
    try:
        rp = parameters.read()
    except FileNotFoundError:
        if not preset_params:
            logger.error("No PARAMETERS file found, and no preset parameters "
                         "passed. Execution will stop.")
            cleanup(tmp_manifest)
            return 2
        rp = rparams.Rparams()
    except Exception:
        logger.error("Exception while reading PARAMETERS file", exc_info=True)
        cleanup(tmp_manifest)
        return 2

    # check if this is going to be a domain search
    domains = False
    if "DOMAIN" in rp.readParams:
        domains = True

    if domains:  # no POSCAR in main folder for domain searches
        slab = None
    elif slab is None:
        poscar_file = Path("POSCAR")
        if not poscar_file.is_file():
            logger.error("POSCAR not found. Stopping execution...")
            cleanup(tmp_manifest)
            return 2

        logger.info("Reading structure from file POSCAR")
        try:
            slab = poscar.read(filename=poscar_file)
        except Exception:
            logger.error("Exception while reading POSCAR", exc_info=True)
            cleanup(tmp_manifest)
            return 2

        if not slab.preprocessed:
            logger.info("The POSCAR file will be processed and overwritten. "
                        "Copying the original POSCAR to POSCAR_user...")
            try:
                shutil.copy2(poscar_file, "POSCAR_user")
            except OSError:
                logger.error("Failed to copy POSCAR to POSCAR_user. Stopping "
                             "execution...")
                cleanup(tmp_manifest)
                return 2
            tmp_manifest.append("POSCAR_user")
    try:
        # interpret the PARAMETERS file
        parameters.interpret(rp, slab=slab, silent=False)
    except (parameters.errors.ParameterNeedsSlabError,
            parameters.errors.SuperfluousParameterError):
        # Domains calculation is the only case in which slab is None
        logger.error('Main PARAMETERS file contains an invalid parameter '
                     'for a multi-domain calculation', exc_info=True)
        cleanup(tmp_manifest)
        return 2
    except parameters.errors.ParameterError:
        logger.error("Exception while reading PARAMETERS file", exc_info=True)
        cleanup(tmp_manifest)
        return 2

    rp.timestamp = timestamp
    rp.manifest = tmp_manifest

    # Load parameter presets, overriding those in PARAMETERS
    preset_params = preset_params or {}
    try:
        rp.update(preset_params)
    except (ValueError, TypeError):
        logger.warning(f"Error applying preset parameters: ", exc_info=True)

    # set logging level
    if 'LOG_LEVEL' in preset_params:
        logger.info(f'Overriding log level to {rp.LOG_LEVEL}.')
    logger.setLevel(rp.LOG_LEVEL)
    logger.debug("PARAMETERS file was read successfully")

    if not domains:
        warn_if_slab_has_atoms_in_multiple_c_cells(slab, rp)
        slab.full_update(rp)   # gets PARAMETERS data into slab
        rp.fileLoaded["POSCAR"] = True

    # set source directory
    try:
        rp.source_dir = get_tensorleed_path(source)
    except (ValueError, FileNotFoundError) as exc:
        logger.warning(f'{exc} This may cause errors.')
        rp.source_dir = Path(source or '').resolve()

    if system_name is None:
        system_name = _get_parent_directory_name()
    rp.systemName = system_name

    # check if halting condition is already in effect:
    if rp.halt >= rp.HALTING:
        logger.info("Halting execution...")
        cleanup(rp.manifest, rp)
        return 0

    rp.updateDerivedParams()
    logger.info(f"ViPErLEED is using TensErLEED version {str(rp.TL_VERSION)}.")

    prerun_clean(rp, log_name)
    exit_code = section_loop(rp, slab)

    # Finalize logging - if not done, will break unit testing
    logger.handlers.clear()
    logging.shutdown()

    return exit_code



def _get_parent_directory_name():
    """Return the name of the directory above the current one."""
    _system_name = Path.cwd().resolve().parent.name
    logger.info('No system name specified. Using name of parent '
                f'directory: {_system_name}')
    return _system_name



"""Section Full-dynamic Optimization."""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2021-10-22'
__license__ = 'GPLv3+'

import copy
import logging
import os
from pathlib import Path
import shutil

import numpy as np
from numpy.polynomial import Polynomial

from viperleed.calc import psgen
from viperleed.calc.files import iofdopt
from viperleed.calc.files import parameters
from viperleed.calc.files import poscar
from viperleed.calc.sections.refcalc import refcalc as section_refcalc
from viperleed.calc.sections.rfactor import rfactor as section_rfactor

logger = logging.getLogger(__name__)


class FullDynamicCalculationError(Exception):
    """Base class for exceptions raised by the full dynamic calculation."""

    def __init__(self, message):
        super().__init__(message)

class FullDynamicOptimizationOutOfBoundsError(FullDynamicCalculationError):
    """Raised when the optimization is out of bounds."""
    def __init__(self, message):
        super().__init__(message)

def get_fd_r(sl, rp, work_dir=Path(), home_dir=Path()):
    """
    Runs reference calculation and r-factor calculation, returns R.

    Parameters
    ----------
    sl : Slab
        Object containing atomic configuration
    rp : Rparams
        Object containing run parameters
    work_dir : pathlike, optional
        The directory to execute the calculations in. The default is ".".
    home_dir : pathlike, optional
        The directory to return to after finishing. The default is the current
        working directory.

    Returns
    -------
    rfactor : float
        The r-factor obtained for the sl, rp combination
    """
    rp.TENSOR_OUTPUT = [0]
    rp.workdir = Path(work_dir).resolve()
    # internally transform theta, phi to within range
    if rp.THETA < 0:
        rp.THETA = abs(rp.THETA)
        rp.PHI += 180
    rp.THETA = min(rp.THETA, 90.)
    rp.PHI = rp.PHI % 360
    # create work directory, go there, execute
    try:
        if work_dir != ".":
            os.makedirs(work_dir, exist_ok=True)
            os.chdir(work_dir)
        for fn in ["BEAMLIST", "PHASESHIFTS"]:
            try:
                shutil.copy2(home_dir / fn, fn)
            except Exception:
                logger.error(f"Error copying {fn} to subfolder.")
                raise
        logger.info("Starting full-dynamic calculation")
        try:
            section_refcalc(sl, rp, parent_dir=home_dir)
        except Exception:
            logger.error("Error running reference calculation")
            raise
        logger.info("Starting R-factor calculation...")
        try:
            rfaclist = section_rfactor(sl, rp, 11)
        except Exception:
            logger.error("Error running rfactor calculation")
            raise
    finally:
        os.chdir(home_dir)
    return rp.last_R, rfaclist


def apply_scaling(sl, rp, which, scale):
    scaling = (scale if char in which else 1 for char in 'abc')
    sl.update_fractional_from_cartesian()                                       # TODO: needed?
    sl.bulkslab.update_fractional_from_cartesian()                              # TODO: needed?
    sl.apply_scaling(*scaling)
    rp.BULK_REPEAT = sl.bulkslab.get_bulk_repeat(rp)


def fd_optimization(sl, rp):
    """
    Runs multiple consecutive reference calculations and r-factor calculations
    optimizing a parameter, as defined by OPTIMIZE in PARAMETERS.

    Parameters
    ----------
    sl : Slab
        Object containing atomic configuration
    rp : Rparams
        Object containing run parameters

    Returns
    -------
    None.
    """

    which = rp.OPTIMIZE["which"]
    if which == "none":
        logger.error("No parameter is defined for optimization. Set OPTIMIZE "
                     "in the PARAMETERS file to define behaviour.")
        rp.setHaltingLevel(2)
        return
    logger.info(f"Starting optimization of {rp.OPTIMIZE['which']}")
    # make sure there's a compiler ready, and we know the number of cores:
    if rp.FORTRAN_COMP[0] == "":
        try:
            rp.getFortranComp()
        except Exception:
            logger.error("No fortran compiler found, cancelling...")
            raise
    rp.updateCores()  # if number of cores is not defined, try to find it
    # check whether step is set; if not, choose default value
    if rp.OPTIMIZE["step"] == 0.:
        if which in ["v0i", "theta"]:
            rp.OPTIMIZE["step"] = 0.5
        elif which == "phi":
            rp.OPTIMIZE["step"] = 2.
        else:   # unit cell size
            rp.OPTIMIZE["step"] = 0.01
        logger.debug("Initial step size undefined, defaulting to {}"
                     .format(rp.OPTIMIZE["step"]))
    if rp.OPTIMIZE["maxstep"] == 0.:
        rp.OPTIMIZE["maxstep"] = 3 * rp.OPTIMIZE["step"]
    if rp.OPTIMIZE["convergence"] == 0.:
        rp.OPTIMIZE["convergence"] = 0.1 * rp.OPTIMIZE["step"]
    rp.OPTIMIZE["maxstep"] = abs(rp.OPTIMIZE["maxstep"])  # always positive
    rp.OPTIMIZE["convergence"] = abs(rp.OPTIMIZE["convergence"])
    # loop over required configurations, copying sl and rp every time
    known_points = np.empty((0, 2))
    # known_points will contain columns [x, r], with x the parameter value
    rfactor_lists = []    # for each point, store r-factors per beam
    tmpdirs = []
    if which == "v0i":
        x0 = rp.V0_IMAG
    elif which == "theta":
        x0 = rp.THETA
    elif which == "phi":
        x0 = rp.PHI
    else:
        x0 = 1.   # geometry: x is a scaling factor



    # optimization loop
    curvature_fail = 0
    parabola = None
    best_explicit_r = None
    while True:
        if rp.STOP:
            break
        # Make sure that there are no duplicate atoms. This is more
        # a safeguard for the future, if we allow also full-dynamic
        # optimization of atoms.
        sl.check_atom_collisions(rp.SYMMETRY_EPS)
        if len(known_points) == 0:
            x = x0   # x will be the value of the parameter under variation
        elif len(known_points) == 1:
            x = x0 + rp.OPTIMIZE["step"]   # first step is given
        elif len(known_points) == 2:  # get third point
            if known_points[1, 1] < known_points[0, 1]:
                x = x0 + 2*rp.OPTIMIZE["step"]
            else:
                x = x0 - rp.OPTIMIZE["step"]
        else:
            parabola = Polynomial.fit(known_points[:, 0], known_points[:, 1],
                                      2)
            coefs = parabola.convert(domain=[-1, 1]).coef
            current_scope = (min(known_points[:, 0]), max(known_points[:, 0]))
            # first check curvature
            if coefs[2] <= 0:  # TODO: some epsilon would be better
                # no good slope - keep adding points on either side
                curvature_fail += 1
                if (len([v for v in known_points[:, 0] if v > x0])
                        > len([v for v in known_points[:, 0] if v < x0])):
                    x = current_scope[0] - (abs(rp.OPTIMIZE["step"])
                                            * curvature_fail)
                else:
                    x = current_scope[1] + (abs(rp.OPTIMIZE["step"])
                                            * curvature_fail)
                logger.info("Current fit parabola is not concave, no minimum "
                            "predicted. Adding data point at {:.4f}".format(x))
            else:
                # try to use minimum
                new_min = -0.5*coefs[1] / coefs[2]
                x = new_min
                # if at the edge, better to step a bit farther again
                if x < current_scope[0]:
                    x = min(x, current_scope[0] - abs(rp.OPTIMIZE["step"]))
                elif x > current_scope[1]:
                    x = max(x, current_scope[1] + abs(rp.OPTIMIZE["step"]))
                # limit to maxstep
                x = min(x, current_scope[1] + rp.OPTIMIZE["maxstep"])
                x = max(x, current_scope[0] - rp.OPTIMIZE["maxstep"])
                if which == "v0i":
                    x = max(0, x)
                if which in "abc":
                    x = max(0.1, x)   # shouldn't happen, just in case
                # check whether we're close to point that is already known
                if any(abs(v - x) < rp.OPTIMIZE["convergence"]
                       for v in known_points[:, 0]):
                    left = [v for v in known_points[:, 0] if v < x]
                    right = [v for v in known_points[:, 0] if v > x]
                    if (not left or
                        all(abs(v - x) < rp.OPTIMIZE["convergence"]
                            for v in left)):     # take step to left
                        x = current_scope[0] - abs(rp.OPTIMIZE["step"])
                    elif (not right or
                          all(abs(v - x) < rp.OPTIMIZE["convergence"]
                              for v in right)):  # take step to right
                        x = current_scope[1] + abs(rp.OPTIMIZE["step"])
                    elif (all(abs(v - x) < rp.OPTIMIZE["convergence"]
                              for v in (max(left), min(right)))
                          and len(known_points) < rp.OPTIMIZE["minpoints"]):
                        # points are very close already, default to adding
                        #   points outside of scope
                        if (len(right) > len(left)):
                            x = current_scope[0] - abs(rp.OPTIMIZE["step"])
                        else:
                            x = current_scope[1] + abs(rp.OPTIMIZE["step"])
                    else:        # take midpoint
                        x = (max(left) + min(right)) / 2
                    # stop if minpoints and convergence
                    if len(known_points) >= rp.OPTIMIZE["minpoints"]:
                        if not (current_scope[0] < x < current_scope[1]):
                            logger.info(
                                "Minimum has converged, but is at the edge of "
                                "the tested range. Proceeding with "
                                "optimization.")
                        else:
                            lowest_dist = min([abs(v - new_min) for v
                                               in known_points[:, 0]])
                            closest_point = [
                                v for v in known_points[:, 0]
                                if abs(v - new_min) == lowest_dist][0]
                            logger.info(
                                "Stopping optimization: Predicted new minimum "
                                f"{new_min:.4f} is within convergence limits to "
                                f"known point {closest_point:.4f}.")
                            break  # within convergence limit
                logger.info("Currently predicting minimum at {:.4f} with R = "
                            "{:.4f}, adding data point at {:.4f}"
                            .format(new_min, parabola(new_min), x))

        # write out results
        if len(known_points) != 0:
            iofdopt.write_fd_opt_csv(known_points, which)
        if len(known_points) > 2:
            iofdopt.write_fd_opt_pdf(known_points, which, parabola=parabola)

        # create test objects tsl, trp and set parameters
        tsl = copy.deepcopy(sl)
        trp = copy.deepcopy(rp)
        if which == "v0i":
            x = max(0, x)
            trp.V0_IMAG = x
        elif which == "theta":
            trp.THETA = x
        elif which == "phi":
            trp.PHI = x
        else:       # geometry: x is a scaling factor for the unit cell
            x = max(0.1, x)
            apply_scaling(tsl, trp, which, x)

        # create subfolder and calculate there
        if isinstance(x, int):
            dname = f"{which}_{x}"
        else:
            dname = f"{which}_{x:.4f}"
        workdir = rp.workdir / dname
        tmpdirs.append(workdir)
        logger.info(f"STARTING CALCULATION AT {which} = {x:.4f}")
        r, rfaclist = get_fd_r(tsl, trp, work_dir=workdir, home_dir=rp.workdir)
        known_points = np.append(known_points, np.array([[x, r]]), 0)
        rfactor_lists.append(rfaclist)
        if not best_explicit_r or r < best_explicit_r[0]:
            best_explicit_r = trp.stored_R["refcalc"]    # has int & frac components
        # decide how to proceed
        if len(known_points) >= rp.OPTIMIZE["maxpoints"]:
            logger.warning(
                "Stopping optimization because the maximum number of data "
                "points has been reached. This may indicate that optimization "
                "did not fully converge.")
            break

    # optimization loop finished; re-fit points, analyze:
    parabola = Polynomial.fit(known_points[:, 0], known_points[:, 1], 2)
    coefs = parabola.convert(domain=[-1, 1]).coef
    new_min = -0.5*coefs[1] / coefs[2]
    logger.info(f"Optimization of {which}: Predicted minimum at "
                f"{new_min:.4f}, R = {parabola(new_min):.4f}")
    current_best = known_points[np.argmin(known_points, 0)[1]]
    rp.stored_R["refcalc"] = best_explicit_r
    if (round(parabola(new_min), 4) > round(current_best[1], 4)
            and current_best[0] != known_points[-1, 0]):
        logger.warning(
            "Fit prediction at {:.4f} is worse than explicit calculation "
            "result at {:.4f} (R = {:.4f}). Outputting best known result "
            "instead of parabolic prediction.".format(new_min, *current_best))
        new_min = current_best[0]

    # output analysis
    iofdopt.write_fd_opt_csv(known_points, which)
    if len(known_points) > 2:
        iofdopt.write_fd_opt_pdf(known_points, which, parabola=parabola)

    # output modified files
    comment = "Found by full-dynamic optimization"
    if which == "v0i":
        rp.V0_IMAG = new_min
        parameters.modify(rp, "V0_IMAG", comment=comment)
    elif which in ("theta", "phi"):
        setattr(rp, which.upper(), new_min)
        if rp.THETA < 0:
            rp.THETA = abs(rp.THETA)
            rp.PHI += 180
        rp.PHI = rp.PHI % 360
        parameters.modify(rp, "BEAM_INCIDENCE", comment=comment)
    else:       # geometry: x is a scaling factor for the unit cell
        apply_scaling(sl, rp, which, new_min)
        if not isinstance(rp.BULK_REPEAT, float) or "c" in which:
            parameters.modify(rp, "BULK_REPEAT", comment=comment)
        poscar.write(sl, filename=f"POSCAR_OUT_{rp.timestamp}", comments="all")

    # fetch I(V) data from all, plot together
    best_rfactors = rfactor_lists[np.argmin(known_points[:, 1])]
    try:
        iofdopt.write_fd_opt_beams_pdf(rp, known_points, which, tmpdirs,
                                       best_rfactors)
    except Exception as exc:
        logger.warning(f"Failed to plot I(V) curves from optimization: {exc}")

    # clean up tmpdirs
    # for path in tmpdirs:
    #     try:
    #         shutil.rmtree(path)
    #     except Exception as e:
    #         logger.warning("Failed to delete temporary directory {}: "
    #                        .format(path) + str(e))



"""Module utils of viperleed.calc.classes.slab.

Contains functions and classes useful for dealing with BaseSlab objects.
"""

__authors__ = (
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-02-22'
__license__ = 'GPLv3+'

import itertools
import numpy as np


def _cycle(sequence, start=0):                                                  # TODO: could it be useful in other places?
    """Return a generator that cycles though `sequence` beginning at start."""
    if sequence:
        start %= len(sequence)
    _cycled = itertools.cycle(sequence)
    return itertools.islice(_cycled, start, None)


def _left_handed(ab_cell):
    """Return whether a 2D unit cell is left-handed."""
    return np.linalg.det(ab_cell) < 0


def _z_distance(atom1, atom2):                                                  # TODO: could it be useful in other places? If yes, then maybe the right place is .atom?
    """Return the distance along z between two atoms."""
    return abs(atom2.cartpos[2] - atom1.cartpos[2])



"""Moule limits of viperleed.calc.classes.rparams.

Defines the limiting values of 'simple' user PARAMETERS. Not-so-simple
parameters, which are defined as their own classes in package special,
also take care of their own default values. This module was originally
(2019-06-13) part of the rparams.py module, refactored by Michele Riva
in Oct 2023.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-10-23'
__license__ = 'GPLv3+'

                                                                                # TODO: fill dict of parameter limits here (e.g. LMAX etc.)
# parameter limits
# either tuple of (min, max) or list of allowed values                          # TODO: allowed would be cleaner as set. It's not great that things are mixed though. Would be better to have a separate global
PARAM_LIMITS = {
    'LMAX': (1, 18),
    'INTPOL_DEG': ['3', '5'],
    }



"""Functions for writing files relevant to the superpos calculation."""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-19'
__license__ = 'GPLv3+'

import logging

import fortranformat as ff
import numpy as np

logger = logging.getLogger(__name__)

from viperleed.calc.lib.version import Version


def writeSuperposInput(sl, rp, config, param_name="PARAM",
                       contrin_name="superpos-CONTRIN",
                       for_error=False, only_vary=None):
    """
    Writes a PARAM and a CONTRIN file as input for the Superpos calculation.
    Returns the contents of CONTRIN, as this will also be needed as an input
    stream for the superpos calculation.

    Parameters
    ----------
    sl : Slab
        The Slab object to be modified.
    rp : Rparams
        The run parameters.
    config : list of int
        Parameter indices to be applied.
    param_name : str, optional
        Alternative name for the PARAM file.
    contrin_name : str, optional
        Alternative name for the superpos-CONTRIN file.
    for_error : bool, optional
        If True, the 'config' parameter will be ignored, and superpos will
        instead be calculated for all variations.
    only_vary : list of SearchPar, optional
        If running for error, pass a list of search parameters that should be
        varied. All other search parameters will be kept at their "centered"
        position.

    Returns
    -------
    str
        Contents of CONTRIN.

    """
    if only_vary is None:
        only_vary = rp.searchpars
    mnfiles = 0
    occupations = []    # occupation per delta file
    deltanames = []     # names of delta files
    surfaceChecks = []  # is that atom at the surface or not?
    indices = []        # vib and geo indices from search, cast to 1D array
    n_var = 1
    n_conc = 1
    if for_error:
        n_var = max([sp.steps for sp in rp.searchpars
                     if sp.mode in ("geo", "vib") and sp in only_vary])
        n_conc = max([sp.steps for sp in rp.searchpars
                      if sp.mode == "occ" and sp in only_vary])
    # determine which atoms are "at the surface"
    surfats = sl.getSurfaceAtoms(rp)
    # make lists to print
    which_atoms = [at for at in rp.search_atlist
                   if [sp for sp in rp.searchpars
                       if sp.atom == at and (sp.non_zero or sp in only_vary)]]
    for at in which_atoms:
        sps = [sp for sp in rp.searchpars if sp.atom == at]
        occpar = [sp for sp in sps if sp.mode == "occ"][0]  # exactly one
        totalocc = np.array([0.] * n_conc)
        for el in at.disp_occ:
            if at in surfats:
                surfaceChecks.append(1)
            else:
                surfaceChecks.append(0)
            if not for_error:
                occind = rp.searchpars.index(occpar)
                o = [at.disp_occ[el][config[occind]-1]]
            elif occpar not in only_vary:
                o = [at.disp_occ[el][occpar.center-1]] * n_conc
            else:
                try:
                    o = [at.disp_occ[el][j] for j in range(n_conc)]
                except IndexError:
                    logger.error("Superpos for error: Inconsistent number of "
                                 "variation steps. Found maximum {} steps,"
                                 " but only {} for {}, element {}."
                                 .format(n_conc, len(at.disp_occ[el]), at, el))
                    rp.setHaltingLevel(2)
                    return ""
            occupations.append(o)
            totalocc += np.array(o)
            pl = [sp for sp in sps if sp.el == el]
            if not pl:
                logger.error('No search parameters found '
                             f'for {at}. Aborting...')
                rp.setHaltingLevel(2)
                return ""
            deltanames.append(pl[0].deltaname)
            mnfiles += 1
            if not for_error or any(sp not in only_vary for sp in pl):
                if el in at.disp_geo:
                    k = el
                else:
                    k = "all"
                ngeo = len(at.disp_geo[k])
                vibpar = [sp for sp in pl if sp.mode == "vib"]
                if len(vibpar) > 0:
                    if not for_error:
                        vibind = config[rp.searchpars.index(vibpar[0])] - 1
                    else:
                        vibind = vibpar[0].center - 1
                else:
                    vibind = 0
                geopar = [sp for sp in pl if sp.mode == "geo"]
                if len(geopar) > 0:
                    if not for_error:
                        geoind = config[rp.searchpars.index(geopar[0])] - 1
                    else:
                        geoind = geopar[0].center - 1
                else:
                    geoind = 0
                # cast vib and geo indices from 2D to 1D:
                indices.append([(ngeo * vibind) + geoind + 1] * n_var)
            else:
                whichpar = max(pl, key=lambda x: x.steps)
                if whichpar.steps == n_var:
                    indices.append(list(range(1, n_var + 1)))
                else:
                    logger.error("Superpos for error: Inconsistent number of "
                                 "variation steps. Found maximum {} steps,"
                                 " but only {} for {}."
                                 .format(n_var, whichpar.steps, at))
                    rp.setHaltingLevel(2)
                    return ""
        vp = [sp for sp in sps if sp.el == "vac"]
        if len(vp) != 0 and np.any(totalocc < 1.0):
            occupations.append(list(1 - totalocc))
            deltanames.append(vp[0].deltaname)
            mnfiles += 1
            indices.append([1] * n_var)
            if at in surfats:
                surfaceChecks.append(1)
            else:
                surfaceChecks.append(0)

    # write PARAM
    param = """C  DIMENSIONS MAY BE CHANGED IN PARAMETER-STATEMENT
C
C  MNFILES: (maximum) number of different files containing amplitude changes
C           to be used as input files for the I(E) calculation
C  MNCONCS: (maximum) number of different combinations of occupation probabilities
C           for the lattice sites in the surface - i.e., weight factors multiplied
C           to the delta amplitudes from each file
C  MNT0:    (maximum) number of beams for which I(E) spectra are calculated
C  MNCSTEP: (maximum) number of parameter combinations (geometry x vibrations)
C           per energy tabulated in an individual delta amplitude file
C  MNATOMS: currently inactive - set to 1
C
"""
    param += "      PARAMETER (MNFILES={})\n".format(mnfiles)
    param += "      PARAMETER (MNCONCS={})\n".format(n_conc)
    param += ("      PARAMETER (MNT0={}, MNCSTEP={}, MNATOMS=1)\n"
              .format(len(rp.ivbeams), rp.mncstep))
    try:
        with open(param_name, "w") as wf:
            wf.write(param)
    except Exception:
        logger.error("Failed to write to PARAM file for superpos")
        raise

    # collect contrin output
    if rp.TL_VERSION < Version('1.7.0'):
        formatter = {'int': ff.FortranRecordWriter('I3'),
                     'occ': ff.FortranRecordWriter('10F7.4'),
                     'var': ff.FortranRecordWriter('I3'),
                     'ndelta': ff.FortranRecordWriter('100I3'),
                     }
    else:
        formatter = {'int': ff.FortranRecordWriter('I4'),
                     'occ': ff.FortranRecordWriter('10F7.4'),
                     'var': ff.FortranRecordWriter('I3'),
                     'ndelta': ff.FortranRecordWriter('1000I5'),
                     }
    contrin = ((formatter['int'].write([mnfiles])
                + formatter['int'].write([0])).ljust(16)
               + "no. of files, VarAll: all possible parameter combinations? "
               "(1/0)\n")
    contrin += (formatter['int'].write([n_conc]).ljust(16)
                + "number of concentration steps\n")
    occ_trans = map(list, zip(*occupations))  # transpose occupations
    for occ in occ_trans:
        contrin += formatter['occ'].write(occ) + "\n"
    for i in range(0, len(deltanames)):
        contrin += deltanames[i].ljust(15) + formatter['var'].write([n_var])
        contrin += (formatter['var'].write([surfaceChecks[i]])
                    + formatter['var'].write([1]) + "  FILENAME,VARIATIONS,"
                    "SURFACE?,FORMATTED? (A15,3I3)\n")
        contrin += formatter['ndelta'].write(indices[i]) + "\n"
    try:
        with open(contrin_name, "w") as wf:
            wf.write(contrin)
    except Exception:
        logger.error("Failed to write " + contrin_name + " file. Execution "
                     "will continue...")

    return contrin



"""Module surface_slab of viperleed.calc.classes.slab.

Defines the SurfaceSlab class, a BaseSlab describing a 2D-periodic
crystal. This is available in viperleed.calc as the Slab alias for
backwards compatibility with the former module-based structure of
slab.py by @fkraushofer (originally created on 2019-06-13).
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Michele Riva (@michele-riva)'
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-02-21'
__license__ = 'GPLv3+'

from collections import Counter
import copy
import itertools
import logging
from operator import itemgetter

import numpy as np
from scipy.spatial import KDTree, distance as sp_distance

from viperleed.calc.classes.atom import Atom
from viperleed.calc.classes.atom_containers import AtomList
from viperleed.calc.files.parameters.errors import (
    InconsistentParameterError
    )
from viperleed.calc.lib import leedbase
from viperleed.calc.lib.base import NonIntegerMatrixError
from viperleed.calc.lib.base import SingularMatrixError
from viperleed.calc.lib.base import add_edges_and_corners
from viperleed.calc.lib.base import collapse
from viperleed.calc.lib.base import collapse_fractional
from viperleed.calc.lib.base import ensure_integer_matrix
from viperleed.calc.lib.base import pairwise
from viperleed.calc.lib.base import parent_name
from viperleed.calc.lib.periodic_table import COVALENT_RADIUS
from viperleed.calc.lib.periodic_table import PERIODIC_TABLE

from .base_slab import BaseSlab
from .bulk_slab import BulkSlab
from .errors import AlreadyMinimalError
from .errors import AtomsTooCloseError
from .errors import EmptySlabError
from .errors import MissingBulkSlabError
from .errors import MissingLayersError
from .errors import NoBulkRepeatError
from .errors import NoVacuumError
from .errors import NotEnoughVacuumError
from .errors import SlabError
from .errors import TooFewLayersError
from .errors import WrongVacuumPositionError

try:
    import ase
except ImportError:
    _HAS_ASE = False
else:
    _HAS_ASE = True


_LOGGER = logging.getLogger(parent_name(__name__))
_MIN_VACUUM = 5.0   # Angstrom
_VACUUM_EPS = 1e-4  # In fractional c coordinates


class SurfaceSlab(BaseSlab):
    """A class representing a semi-infinite solid.

    Contains unit cell, element information and atom coordinates.
    Also has a variety of convenience functions for manipulating
    and updating the atoms. Slabs can be created from an ase.Atoms
    object using the `from_ase` class method. Another common way
    is using `viperleed.calc.files.poscar.read`.

    Attributes
    ----------
    ucell : np.array
        The unit cell as vectors a, b, c (columns)
    poscar_scaling : float
        The original scaling factor from POSCAR
    elements : tuple of str
        Element labels as read from POSCAR
    chemelem : set of str
        Chemical elements in the slab, including from `ELEMENT_MIX`
    n_per_elem : dict {str: int}
        The number of atoms per POSCAR element.
    atlist : AtomList
        List of all atoms in the slab.
    layers : tuple of Layer
        Each `layer` is a composite of sublayers, as in TensErLEED
    sublayers : tuple of SubLayer
        Each SubLayer contains atoms of equal element and Z coordinate
    sitelist : list of Sitetype
        List of distinct sites as Sitetype, storing information
        on vibration and concentration
    ucell_mod : list of tuples (str, numpy.ndarray)
        Stored modifications made to the unit cell; each is a tuple
        of (type, array), where type is 'lmul', 'rmul', 'add', or
        'c_shift'.
    topat_ori_z : float
        Stores the original position of the topmost atom in Cartesian
        coordinates
    celltype : str                                                              # TODO: would be nicer with an Enum
        Unit-cell shape as string. Values: 'oblique', 'rhombic',
        'rectangular', 'square', 'hexagonal'
    planegroup : str
        Symmetry group of the slab. May be reduced by the user
        relative to `foundplanegroup`.
    foundplanegroup : str
        Highest symmetry found. Doesn't get modified when user
        reduces symmetry manually.
    orisymplane : SymPlane
        Only stored if the `planegroup` is ambiguous as to which unit
        vector the symmetry plane at the origin is parallel to
    linklists : list of list of Atom
        List of lists of atoms which are linked by a symmetry operation
    displists : list of list of Atom
        List of lists of atoms which are displaced together. This
        differs from `linklists` in that while `linklists` stores
        the linking based on the current symmetry, the `displists`
        store actual displacement linking based on the symmetry set
        at the time the displacement is assigned.
    deltas_initialized : bool
        Set by Rparams.generateSearchPars
    preprocessed : bool
        True if the POSCAR that this slab was read from had the
        'Plane group = XY' comment in the header, indicating that
        it was processed by viperleed.calc before.
    symbaseslab : Slab or None
        Slab with the smallest in-plane unit-cell area that shows
        the full symmetry of the slab.
    bulkslab : Slab or None
        Slab object containing only bulk layers
    """

    def __init__(self):
        """Initialize instance."""
        super().__init__()
        self.displists = []
        self.preprocessed = False
        self.deltas_initialized = False
        self.symbaseslab = None

    @property
    def is_bulk(self):
        """Return whether this is a bulk slab."""
        return False

    @property
    def thickness(self):
        """Return the thickness of this slab along z."""
        top_atom_z = min(at.cartpos[2] for at in self)                          # TODO: swap min and max with .cartpos[2] -- Issue #174
        bottom_atom_z = max(at.cartpos[2] for at in self)
        return abs(top_atom_z - bottom_atom_z)

    @property
    def vacuum_gap(self):
        """Return the z distance that does not not contain atoms."""
        return self.c_vector[2] - self.thickness

    @classmethod
    def from_ase(cls, ase_atoms, sort_elements=True):
        """Return a slab initialized from an ase.Atoms object.

        Parameters
        ----------
        ase_atoms : ase.Atoms
            The Atomic-simulation-environment Atoms object to
            initialize from.
        sort_elements : bool, optional
            Whether chemical elements should be sorted in
            alphabetical order before storing. Useful to
            prevent incorrect ordering when an existing
            PHASESHIFTS file is used. Default is True.

        Returns
        -------
        slab : Slab
            The Slab, initialized from ase_atoms. May be empty
            if the ase module is not available.

        Raises
        ------
        ModuleNotFoundError
            If the ase module is not available.
        TypeError
            If `ase_atoms` is not an ase.Atoms object.
        """
        if not _HAS_ASE:
            _err_msg = ('Slab creation from ase.Atoms not '
                        'supported: Module ase not found')
            _LOGGER.error(_err_msg)
            raise ModuleNotFoundError(_err_msg, name='ase', path=__file__)
        if not isinstance(ase_atoms, ase.Atoms):
            raise TypeError(f'{cls.__name__}.from_ase requires an ase.Atoms '
                            f'object, found {type(ase_atoms)!r} instead.')
        self = cls()
        self.ucell = ase_atoms.cell.copy().T
        elems = ase_atoms.symbols
        n_per_el = dict(Counter(elems))
        if sort_elements:
            n_per_el = {k: n_per_el[k] for k in sorted(n_per_el.keys())}
        self.n_per_elem = n_per_el
        for elem, pos in zip(elems, ase_atoms.get_scaled_positions()):
            self.atlist.append(Atom(elem, pos, self.n_atoms + 1, self))
        self.update_cartesian_from_fractional()
        return self

    @classmethod
    def from_slab(cls, other):
        """Return a `cls` instance with attributes deep-copied from `other`.

        Parameters
        ----------
        other : BaseSlab
            The slab whose attributes are to be copied.

        Returns
        -------
        new_slab : SurfaceSlab
            A new slab instance with attributes copied from `other`.
            `new_slab` will not have .layers and .sublayers defined
            if other.is_bulk. This is to prevent `new_slab` from
            having too many (i.e., more than RParams.N_BULK_LAYERS)
            bulk layers.
        """
        new_slab = super().from_slab(other)
        if other.is_bulk:
            new_slab.layers = ()
            new_slab.sublayers = ()
        return new_slab

    # Used only in ensure_minimal_bulk_ab_cell
    def _change_bulk_cell(self, rpars, new_ab_cell):
        """Assign a new 2D unit cell to the bulk, if possible.

        The new cell is applied only if it gives an acceptable
        `SUPERLATTICE` matrix. Otherwise exceptions are raised.

        Parameters
        ----------
        rpars : Rparams
            The PARAMETERS to be used and updated. Attributes accessed:
            (read/write) SUPERLATTICE; (read) BULK_REPEAT, SYMMETRY_EPS,
            superlattice_defined.
        new_ab_cell : Sequence
            Shape (2, 2), representing the new 2D unit cell in
            Cartesian coordinates. The new unit vectors should
            be rows. A tentative SUPERCELL is calculated from it.

        Raises
        ------
        NonIntegerMatrixError
            If `new_ab_cell` gives a superlattice matrix that is not
            integer-valued.
        calc.files.parameters.errors.InconsistentParameterError
            If `new_ab_cell` gives a different superlattice matrix
            than the one in `rpars`.
        """
        # Calculate new SUPERLATTICE matrix
        superlattice = self.ab_cell.T.dot(np.linalg.inv(new_ab_cell))
        try:
            superlattice = ensure_integer_matrix(superlattice, eps=5e-2)
        except NonIntegerMatrixError:
            raise NonIntegerMatrixError(
                'Automatically detected bulk SUPERLATTICE is '
                f'not integer-valued: \n{superlattice}'                         # TODO: guilib array formatter
                ) from None
        if (rpars.superlattice_defined
                and not np.allclose(superlattice, rpars.SUPERLATTICE)):
            raise InconsistentParameterError(
                'Automatically detected minimum-area bulk unit cell differs '
                'from the cell defined by the SUPERLATTICE parameter. '
                'Consider changing the SUPERLATTICE parameter. Found matrix: '
                f'\n{superlattice.astype(int)}'                                 # TODO: guilib array formatter
                )
        rpars.SUPERLATTICE = superlattice

        if not self.bulkslab:
            raise MissingBulkSlabError
        # Notice that there is no need to re-center along bulk c, as
        # we're only modifying the in-plane cell. There's also no
        # reason to modify the c coordinates, hence z_periodic=False
        self.bulkslab.apply_bulk_cell_reduction(rpars.SYMMETRY_EPS,
                                                epsz=rpars.SYMMETRY_EPS.z,
                                                new_ab_cell=new_ab_cell,
                                                recenter=False,
                                                z_periodic=False)

    def check_atom_collisions(self, eps=0.05):
        """Raise if any pairs of Atoms are too close to one another.

        Parameters
        ----------
        eps : float, optional
            Minimum Cartesian distance (Angstrom) for Atoms to be
            considered colliding. Default is 0.05.

        Raises
        ------
        MissingLayersError
            If this method is called before this Slab has layers.
        AtomsTooCloseError
            If any pair of atoms, including 2D replicas, are too
            close to one another.
        """
        if not self.layers:
            raise MissingLayersError('Cannot check_atom_collisions '
                                     'without layers')

        ucell = self.ucell.T
        ucell_inv = np.linalg.inv(ucell)
        releps = eps / np.linalg.norm(self.ab_cell.T, axis=1)

        # We will check atoms including 2D replicas. We should take
        # atoms from two adjacent layers if the layers are close
        # enough (most of the times this will not be the case).
        layers = (*self.layers, None)
        for upper_layer, lower_layer in pairwise(layers):
            cartesians = [at.cartpos for at in upper_layer]
            try:
                layer_dist = abs(lower_layer.cartori[2] - upper_layer.cartbotz)
            except AttributeError:  # lower_layer is None
                layer_dist = 2*eps  # Just to skip the next check
            if layer_dist < eps:
                cartesians.extend(at.cartpos for at in lower_layer)
            cartesians = np.array(cartesians)
            cartesians[:, 2] = self.topat_ori_z - cartesians[:, 2]              # TODO: .cartpos[2]. We can entirely skip this when we flip -- Issue #174
            cartesians, fractionals = collapse(cartesians, ucell, ucell_inv)
            cartesians, _ = add_edges_and_corners(cartesians, fractionals,
                                                  releps, ucell)
            too_close = sp_distance.pdist(cartesians, 'euclidean') <= eps
            if any(too_close):
                raise AtomsTooCloseError(
                    f'{type(self).__name__} contains atoms closer than {eps} '
                    'angstrom. This is problematic for a LEED calculation. '
                    'Are there duplicate atoms? Remove duplicates, e.g., '
                    f'via slab.remove_duplicate_atoms(eps={eps}, epsz={eps})'
                    )

    def check_vacuum_gap(self):
        """Complain if there's not enough vacuum or it is not at the top.

        An interatomic z distance of at least 5 angstrom is considered
        an acceptable vacuum gap if it is between the topmost atom and
        the (periodic replica of the) bottommost atom.

        This slab is modified only if there is a large-enough vacuum
        gap very close to the top, i.e., if the atom right above the
        vacuum gap is closer to c=1 than 1e-4.

        Raises
        ------
        NoVacuumError
            If the vacuum gap found is extremely small, i.e., if
            the top and bottom atoms are less than 2e-4 apart along
            the c vector.
        NotEnoughVacuumError
            If a small vacuum gap is found (< 5 angstrom).
        WrongVacuumPositionError
            If there is enough vacuum, but it is somewhere in
            the middle of the slab rather than at the top.
        """
        # Work with a copy not to mess around sorting and positions.
        # This will also become the vacuum-fixed version of self.
        self_copy = copy.deepcopy(self)

        # Ensure the c fractional coordinates are in [0, 1).
        # Also keep the Cartesians up to date, as we will look
        # for the pair of atoms that are furthest along z.
        self_copy.collapse_cartesian_coordinates(update_origin=True)
        self_copy.sort_by_z()  # From bottom to top

        # Collect all the Cartesian coordinates. Also include
        # a periodic replica of the bottom-most atom at the top
        z_cartpos = [self_copy.topat_ori_z - at.cartpos[2] for at in self_copy]
        replica_z = self_copy.topat_ori_z - self_copy.atlist[0].cartpos[2]        # TODO: .cartpos[2] -- Issue #174
        replica_z += self_copy.c_vector[2]  # Don't care about in-plane
        z_cartpos.append(replica_z)

        # Notice that taking the max on both the distance and the index
        # ensures that ind_above is always the largest index (i.e., the
        # highest atom) that realizes a certain vacuum gap. This is
        # also the reason for the round(..., 5): avoids float errors
        vacuum_gap, ind_above = max(
            ((round(c_above - c_below, 5), i+1)
             for i, (c_below, c_above) in enumerate(pairwise(z_cartpos)))
            )
        already_at_top = ind_above == self.n_atoms
        if vacuum_gap >= _MIN_VACUUM:
            bottom_atom = self_copy.atlist[0 if already_at_top else ind_above]
            info_msg = (
                f'The slab has a {vacuum_gap:.2f} A vacuum gap above '
                f'{self_copy.atlist[ind_above-1]}, which is not at the top. '
                )
            self._move_vacuum_to_the_top(bottom_atom, already_at_top,
                                         self_copy, info_msg)
            return

        # Not enough vacuum. Assume top/bottom atoms are at top/bottom
        current_vacuum = (z_cartpos[-1]     # Replica of bottom atom
                          - z_cartpos[-2])  # The actual top one
        self._complain_about_small_vacuum(self_copy, current_vacuum)

    @staticmethod
    def _complain_about_small_vacuum(fixed_slab, current_vacuum):
        """Raise VacumError(s) for a slab with too little vacuum."""
        # Assume that the top/bottom atoms are the actual top/bottom.
        # Remember whether the gap is very very small.
        top_atom = fixed_slab.atlist[-1]
        bot_atom = fixed_slab.atlist[0]
        no_vacuum = (bot_atom.pos[2] - top_atom.pos[2]) % 1.0 < 2*_VACUUM_EPS

        # Produce a fixed version: Expand c vector, then ensure
        # that the bottom atom is at least at c==_VACUUM_EPS. Also,
        # give a bit more vacuum than needed to avoid floating-point
        # precision issues. We don't fixed_slab to cause complaints.
        extra_vacuum = _MIN_VACUUM + _VACUUM_EPS - current_vacuum
        c_extra = fixed_slab.c_vector * extra_vacuum / fixed_slab.c_vector[2]
        fixed_slab.c_vector[:] += c_extra
        fixed_slab.update_fractional_from_cartesian()
        if bot_atom.pos[2] < _VACUUM_EPS:
            fixed_slab.translate_atoms_c(_VACUUM_EPS - bot_atom.pos[2])
        err_msg = (
            f'The slab has a {current_vacuum:.2f} A vacuum gap, which is '
            f'smaller than the minimum ({_MIN_VACUUM} A). Will assume '
            f'that {bot_atom} is at the bottom and {top_atom} at the top'
            )
        exc = NoVacuumError if no_vacuum else NotEnoughVacuumError
        raise exc(err_msg, fixed_slab)

    def _move_vacuum_to_the_top(self, bottom_atom, already_at_top,
                                fixed_slab, info_msg):
        """Move atoms along c, if possible, to have vacuum at the top.

        After a successful call to this method, self has its bottom
        atom at c >= 1e-4. This method **does not** check that the
        vacuum gap is large enough!

        Parameters
        ----------
        bottom_atom : Atom
            The atom that was detected as the one right above vacuum.
        already_at_top : bool
            Whether the vacuum gap of self is already in the top
            part of the slab, i.e., there is no atom above it.
        fixed_slab : SurfaceSlab
            A copy of self to be used for producing a vacuum-
            corrected version. Used for error reporting.
        info_msg : str
            Partial error message to be reported in case this slab
            cannot be automatically corrected (i.e., the vacuum is
            somewhere in the middle).

        Raises
        ------
        WrongVacuumPositionError
            If the vacuum gap is somewhere in the middle of self.
        """
        c_shift = collapse_fractional(-bottom_atom.pos[2], eps=_VACUUM_EPS)
        if abs(c_shift) > _VACUUM_EPS and already_at_top:
            # Vacuum is already at the top, and the bottom atom is
            # far enough from the bottom edge. Leave all as is.
            return

        info_msg += f' The bottom atom is {bottom_atom}, at c={-c_shift:.5f}'
        if abs(c_shift) < _VACUUM_EPS:
            # Vacuum is close to the top, but the bottom atom is very
            # close to the c edge. Move it up, at c == _VACUUM_EPS
            c_shift += _VACUUM_EPS
            _LOGGER.info(info_msg + '. Atoms were shifted along c (by '
                         f'{c_shift:.5f}) so that {bottom_atom} is at the '
                         f'bottom (at c={_VACUUM_EPS:.5f}). PARAMETERS (e.g., '
                         'LAYER_CUTS, BULK_LIKE_BELOW, BULK_REPEAT, ...) '
                         'may need adjustment.')
            self.translate_atoms_c(c_shift)
            return

        # The vacuum is somewhere in the middle. Complain.
        fixed_slab.translate_atoms_c(c_shift + _VACUUM_EPS)
        raise WrongVacuumPositionError(info_msg, fixed_slab)

    def detect_bulk(self, rpars, second_cut_min_spacing=1.2):
        """Determine the minimal bulk repeat vector from BULK_LIKE_BELOW.

        Notice that this method modifies `rpars.BULK_REPEAT` and
        `rpars.SUPERLATTICE` so they contain the detected (minimal)
        bulk-repeat vector and the superlattice matrix. It also
        modifies `rpars.LAYER_CUTS` and `rpars.N_BULK_LAYERS` to
        reflect the newly detected bulk. Finally, this slab will
        have its `.layers` and `bulkslab` set accordingly.

        Parameters
        ----------
        rpars : Rparams
            Run parameters object. Attributes accessed:
            (read)  BULK_LIKE_BELOW, SYMMETRY_EPS,
                    SUPERLATTICE, superlattice_defined
            (write) BULK_REPEAT, LAYER_CUTS, N_BULK_LAYERS,
                    SUPERLATTICE.
        second_cut_min_spacing : float, optional
            Minimal z spacing in Angstrom between sublayers of the
            detected bulk unit to make an additional cut through
            the bulk. If no two sublayers are further apart then
            this threshold, the bulk will be treated as a single
            layer. Default is 1.2.

        Returns
        -------
        bulk_cuts : list of float
            Layer cuts for self that generate the bulk.
        bulk_dist : float
            The largest distance found between bulk sublayers. The
            second bulk cut position is placed there if `bulk_dist`
            is larger than `second_cut_min_spacing`.

        Raises
        ------
        ValueError
            If no `BULK_LIKE_BELOW` was defined in `rpars` or if
            `BULK_REPEAT` is already defined.
        NoBulkRepeatError
            If no repeat vector is found.
        """
        if rpars.BULK_LIKE_BELOW <= 0:
            raise ValueError(f'{type(self).__name__}.detect_bulk: rpars '
                             'must have a positive BULK_LIKE_BELOW defined')
        if rpars.BULK_REPEAT is not None:
            raise ValueError(f'{type(self).__name__}.detect_bulk: rpars '
                             'already has a BULK_REPEAT defined')

        # Work with a temporary copy of this slab to mess up layers,
        # and a temporary copy of rpars to modify LAYER_CUTS, and
        # N_BULK_LAYERS. This copy will also store the correct
        # BULK_REPEAT and SUPERLATTICE till we're sure that the
        # procedure is successful
        self_copy = copy.deepcopy(self)
        rpars_copy = copy.deepcopy(rpars)
        rpars_copy.LAYER_CUTS.update_from_sequence([rpars.BULK_LIKE_BELOW])
        rpars_copy.N_BULK_LAYERS = 1
        self_copy.create_layers(rpars_copy)

        # Make sure that the bottom-most atom is at c=0. This is
        # important to prevent the topmost bulk atoms from being
        # back-folded to the bottom when creating the 'thick' bulk
        # slab below.
        bottom_atom_c = self_copy.remove_vacuum_at_bottom(rpars_copy)

        # Create a pseudo-bulk slab to determine the correct repeat
        # c vector: the c vector now is very likely to be wrong (as
        # it is just chopped off from the one of self). Notice that
        # here we should not re-center the coordinates, as we know
        # that the bulk repeat is incorrect.
        self_copy.make_bulk_slab(rpars_copy, recenter=False)

        # Reduce in-plane bulk. This also updates .SUPERLATTICE
        self_copy.ensure_minimal_bulk_ab_cell(rpars_copy)

        # Detect new c vector (z_periodic=False because current c is
        # the one of self, and is most likely wrong), and collapse cell
        # in the process to get the proper bulk slab. Notice that the
        # next call also saves the new c vector in .BULK_REPEAT
        try:
            self_copy.bulkslab.ensure_minimal_c_vector(rpars_copy,
                                                       z_periodic=False)
        except AlreadyMinimalError as exc:
            _LOGGER.error('Automatic bulk detection failed: Found no bulk '
                          'repeat vector below the specified cut-off.')
            raise NoBulkRepeatError('Failed to detect '
                                    'bulk repeat vector') from exc

        # Identify the cut positions that give bulk layers
        # pylint: disable-next=protected-access
        bulk_cuts, bulk_dist = self_copy._get_bulk_cuts(rpars.SYMMETRY_EPS.z,
                                                        second_cut_min_spacing)
        bulk_cuts = [b + bottom_atom_c for b in bulk_cuts]

        # Finally, update self and rpars with the new information
        rpars.BULK_REPEAT = rpars_copy.BULK_REPEAT
        rpars.SUPERLATTICE = rpars_copy.SUPERLATTICE
        rpars.N_BULK_LAYERS = len(bulk_cuts)
        new_cuts = self.create_layers(rpars, bulk_cuts=bulk_cuts)               # TODO: Issue #121
        rpars.LAYER_CUTS.update_from_sequence(new_cuts)
        self.make_bulk_slab(rpars)
        return bulk_cuts, bulk_dist

    def _get_bulk_cuts(self, epsz, second_cut_min_spacing):
        """Return cut positions for self to give bulk layers."""
        # Calculate cut plane (fractional) between bulk and non-bulk
        # portion of self: will take atoms in the bottommost non-bulk
        # layer, and those in the topmost bulk layer. However, take
        # into account that we may not yet have self.layers defined
        # (nor their bulk or non-bulk nature)
        frac_atoms_z = [at.pos[2] for at in self]

        bulk_height = abs(self.bulkslab.c_vector[2])
        frac_bulk_thickness = bulk_height / abs(self.c_vector[2])
        frac_lowest_pos = min(frac_atoms_z)
        frac_bulk_onset = (frac_lowest_pos + frac_bulk_thickness
                           - (epsz / self.c_vector[2]))
        slab_cuts = [(max(f for f in frac_atoms_z if f < frac_bulk_onset)
                     + min(f for f in frac_atoms_z if f > frac_bulk_onset))
                     / 2]

        if self.bulkslab.n_sublayers == 1:
            return slab_cuts, 0.

        # Look for second cut, where bulk sublayers are furthest apart
        lays_and_distances = (
            (lay_above, abs(lay_above.cartbotz - lay_below.cartbotz))
            for lay_above, lay_below in pairwise(self.bulkslab.sublayers)
            )
        lay_above_cut, maxdist = max(lays_and_distances, key=itemgetter(1))
        if maxdist >= second_cut_min_spacing:
            # Get fractional position, in bulk cell, of the cut
            bulk_frac_cut_from_lowest = (
                lay_above_cut.pos[2]
                - (0.5 * maxdist) / bulk_height
                - min(lay.pos[2] for lay in self.bulkslab.sublayers)
                )
            slab_cuts.append(
                frac_lowest_pos
                + bulk_frac_cut_from_lowest * frac_bulk_thickness
                )
        return slab_cuts, maxdist

    def ensure_minimal_bulk_ab_cell(self, rpars, warn_convention=False):
        """Make sure the `bulkslab` has the smallest possible in-plane cell.

        Parameters
        ----------
        rpars : Rparams
            The PARAMETERS to be used and updated. Attributes accessed:
            (read/write) SUPERLATTICE; (read) BULK_REPEAT, SYMMETRY_EPS,
            superlattice_defined; setHaltingLevel method
        warn_convention : bool, optional
            Whether warnings should be logged in case the reduced cell
            could not be made to follow standard conventions.

        Raises
        ------
        NonIntegerMatrixError
            If reduction to a minimal cell would give a SUPERLATTICE
            matrix with non-integer values.
        calc.files.parameters.errors.InconsistentParameterError
            If reduction to a minimal cell would give a SUPERLATTICE
            different from the one in `rpars`.
        """
        bulk = self.bulkslab or self.make_bulk_slab(rpars)
        eps = rpars.SYMMETRY_EPS
        try:
            mincell = bulk.get_minimal_ab_cell(eps, eps.z, warn_convention)
        except AlreadyMinimalError:
            return
        self._change_bulk_cell(rpars, mincell)

    def get_bulk_repeat(self, rpars, only_z_distance=False):
        """Return the bulk repeat vector (with positive z).

        Notice that this method does **not attempt to identify an
        unknown bulk-repeat vector**. If that's what you're after,
        use `identify_bulk_repeat()` instead.

        Parameters
        ----------
        rpars : Rparams
            The PARAMETERS to be interpreted.
        only_z_distance : bool, optional
            Whether a distance in the direction perpendicular to the
            surface (i.e., not necessarily along the c axis) should
            be returned rather than a full vector. This is ignored
            if `rpars.BULK_REPEAT` is a vector. Default is False.

        Returns
        -------
        bulk_repeat_vector : numpy.ndarray or float
            Bulk repeat vector pointing from the bulk to the surface,
            or its component along z. If `rpars.BULK_REPEAT` is a
            vector, a copy is returned. Otherwise, the vector is taken
            to be parallel to the c axis of this slab. Its length is
            calculated from a z distance (i.e., perpendicular to the
            surface). The z distance is taken as either the value of
            `rpars.BULK_REPEAT` (if it is not-None) or the z distance
            between the bottommost points of the lowest bulk and lowest
            non-bulk layers.

        Raises
        ------
        TooFewLayersError
            If rpars.BULK_REPEAT is None, and this slab has either
            no bulk layers or no non-bulk layers.
        """
        if isinstance(rpars.BULK_REPEAT, np.ndarray):
            bulkc = rpars.BULK_REPEAT.copy()
            if bulkc[2] < 0:
                bulkc *= -1
            return bulkc

        zdiff = (rpars.BULK_REPEAT if rpars.BULK_REPEAT is not None
                 else self._get_bulk_to_non_bulk_distance())
        if only_z_distance:
            return zdiff
        return self.c_vector * zdiff / self.c_vector[2]

    def _get_bulk_to_non_bulk_distance(self):
        """Return the z distance from bottommost bulk and non-bulk layers."""
        bulk_layers = self.bulk_layers
        non_bulk_layers = self.non_bulk_layers
        if not bulk_layers:
            raise TooFewLayersError(
                f'{type(self).__name__} has no bulk layers. '
                'Did you forget to .create_layers(rpars)?'
                )
        if not non_bulk_layers:
            raise TooFewLayersError(f'{type(self).__name__} has only bulk '
                                    'layers. Check LAYER_CUTS or explicitly '
                                    'define a BULK_REPEAT.')
        # Use the distance between the bottommost bulk layer and the
        # bottommost non-bulk layer, i.e., the current 'thickness of
        # bulk', plus the gap between 'bulk' and 'non-bulk' parts.
        return bulk_layers[-1].cartbotz - non_bulk_layers[-1].cartbotz          # TODO: Issue #174?

    def get_nearest_neighbours(self):
        """Return the nearest-neighbour distance for all atoms.

        Returns
        -------
        nn_distances : dict
            Form is {`atom`: `nn_distance`}, with `nn_distance`
            the shortest distance between `atom` and all its
            neighbours, taking into account the periodicity of
            the slab in the direction parallel to the surface.
        """
        # Work with a supercell to account for atoms at edges/corners.
        # Decide based on unit vector lengths how many times we want
        # to repeat in each direction: the more the unit vectors are
        # mismatched the more repetitions.                                      # TODO: there may be a better way taking angle into account as well.
        *ab_cell, _ = self.ucell.T  # ab_cell is 2x3
        ab_norms = np.linalg.norm(ab_cell, axis=1)
        repeats = np.ceil(ab_norms.max() / ab_norms)  # '+' direction
        repeats = 2*repeats + 1           # '+', '-', and centre cell
        supercell = self.make_supercell(np.diag(repeats))

        # For NN query use KDTree from scipy.spatial
        atom_coords_supercell = [atom.cartpos for atom in supercell]
        tree = KDTree(atom_coords_supercell)

        # Coordinates of all atoms in the centre cell:
        center_cell_offset = (repeats + 1) / 2  # Indices only
        center_cell_offset = center_cell_offset.dot(ab_cell)
        atom_coords_center = [atom.cartpos + center_cell_offset
                              for atom in self]

        # Get all distances. Use k==2 as k==1 would just return
        # a bunch of zeros since all atom_coords_center are a
        # subset of atom_coords_supercell: their first nearest
        # neighbour is themselves
        distances, _ = tree.query(atom_coords_center, k=2)
        distances = distances[:, 1]  # First column is zeros
        return dict(zip(self, distances))

    def getSurfaceAtoms(self, rp):
        """Checks which atoms are 'at the surface', returns them as a set."""

        _PTL = set(el.lower() for el in PERIODIC_TABLE)
        _RADII = COVALENT_RADIUS

        atoms = copy.deepcopy(self.atlist)
        # run from top to bottom of slab
        atoms.sort(key=lambda atom: -atom.pos[2])
        covered = set()
        surfats = set()
        for atom in atoms:                                                      # TODO: make this radius into an Atom property
            if atom.el in rp.ELEMENT_MIX:
                # radius as weighted average
                totalocc = 0.0
                r = 0.0
                for chemel in rp.ELEMENT_MIX[atom.el]:
                    if chemel.lower() in _PTL:
                        if chemel in atom.site.occ:
                            r += (_RADII[chemel.capitalize()]
                                  * atom.site.occ[chemel])
                            totalocc += atom.site.occ[chemel]
                    else:
                        _LOGGER.error(
                            'Error identifying surface atoms: Could '
                            f'not identify {chemel} as a chemical element.')
                        rp.setHaltingLevel(2)
                        return []
                if totalocc == 0:
                    _LOGGER.error('Unexpected point encountered in '
                                  'generateSearchInput: GS001')
                else:
                    r /= totalocc
            else:
                # radius of atom
                chemical_element = rp.ELEMENT_RENAME.get(atom.el, atom.el)
                try:
                    r = _RADII[chemical_element.capitalize()]
                except KeyError:
                    _LOGGER.error('Error identifying surface atoms: Could not '
                                  f'identify {atom.el} as a chemical element.')
                    rp.setHaltingLevel(2)
                    return []
            r *= 1.2
            surfats.update(a for a in self
                           if a.pos[2] >= atom.pos[2] and a not in covered)
            covered.update(
                a for a in self
                if a.pos[2] < atom.pos[2] and a.is_same_xy(atom, eps=r)
                )
            if len(covered) + len(surfats) >= len(atoms):
                break   # that's all of them
        return surfats

    def has_atoms_in_multiple_c_cells(self):
        """Return whether atoms appear in multiple cells along c.

        It makes sense to call this method only before the
        first call to collapse_cartesian_coordinates() or
        collapse_fractional_coordinates(). After that, this
        method will always return True.

        Returns
        -------
        has_multiple_cells : bool
            Whether the fractional coordinates of the atoms in this
            slab are found in more than one base unit cell along c.
            Atoms very close (<1e-4) to cell boundaries are not taken
            into consideration for the test.

        Raises
        ------
        EmptySlabError
            If this method is called before this slab has any atom.
        """
        if not self.n_atoms:
            raise EmptySlabError
        cells = (at.pos[2] // 1
                 for at in self
                 if _VACUUM_EPS < at.pos[2] % 1.0 < 1 - _VACUUM_EPS)
        try:
            first_cell = next(cells)
        except StopIteration:  # All atoms close to the edges
            return False
        return any(cell != first_cell for cell in cells)

    def identify_bulk_repeat(self, eps, epsz=None):
        """Find a repeat vector for which the bulk matches the slab above.

        Notice that this may not be the shortest of such vectors: the
        vector returned is the the shortest among those that bring
        `bulkslab` to match the bottommost part of the non-bulk part
        of this slab. Should `bulkslab` be 'too thick' (e.g., the
        LAYER_CUTS select composite layers with an exaggerated number
        of sublayers), the vector returned will not be the shortest
        possible. If you are after the absolutely shortest vector, use
        `slab.bulkslab.get_minimal_c_vector(eps, epsz)`, or
        `slab.bulkslab.ensure_minimal_c_vector(rpars)`.

        Parameters
        ----------
        eps : float
            Cartesian tolerance for in-plane comparisons (Angstrom)
        epsz : float or None, optional
            Cartesian tolerance for comparisons along the direction
            perpendicular to the surface. If not given or None, take
            the same value as `eps`. This argument is used only if
            this slab or its `bulkslab` do not have yet `sublayers`.
            Default is None.

        Returns
        -------
        repeat_vec : numpy.ndarray
            Repeat vector in Cartesian coordinates, with (x, y) in the
            surface plane, and z directed from the solid towards the
            surface, i.e., opposite to the usual LEED convention.

        Raises
        ------
        MissingBulkSlabError
            If this slab has no `bulkslab`.
        NoBulkRepeatError
            If a bulk repeat vector could not be found.
        """
        if self.bulkslab is None:
            raise MissingBulkSlabError(
                f'{type(self).__name__}.identify_bulk_repeat: missing '
                'bulkslab. Call make_bulk_slab, then try again'
                )
        if epsz is None:
            epsz = eps
        if not self.sublayers:
            self.create_sublayers(epsz)
        if not self.bulkslab.sublayers:
            self.bulkslab.create_sublayers(epsz)

        n_bulk_lay = self.bulkslab.n_sublayers
        if self.n_sublayers < 2 * n_bulk_lay:
            raise NoBulkRepeatError(
                f'{type(self).__name__}.identify_bulk_repeat: failed. '
                f'Too few sublayers in slab ({self.n_sublayers}) with '
                f'respect to bulkslab ({n_bulk_lay}). Should be at least '
                f'{2*n_bulk_lay}'
                )

        # Pick the region of the slab that should be considered for
        # comparison: from the topmost bulk layer all the way down.
        # Notice that we take the topmost bulk from self rather than
        # from self.bulkslab, as that one may have a different frame
        # if it was re-centred along z when creating it.
        z_range = (
            self.sublayers[-n_bulk_lay].cartbotz,  # topmost bulk               # TODO: .cartpos[2] -- Issue #174
            self.sublayers[-1].cartbotz            # bottommost
            )

        # Construct candidate repeat vectors as all those connecting
        # the first non-bulk layer to all atoms of the bottommost layer
        first_non_bulk_layer = self.sublayers[-1-n_bulk_lay]
        if first_non_bulk_layer.element != self.sublayers[-1].element:
            # Can't detect a repeat vector if elements don't match
            raise NoBulkRepeatError(
                f'{type(self).__name__}.identify_bulk_repeat: failed. '
                'Chemical elements mismatched in first non-bulk sublayer '
                f'({first_non_bulk_layer.num}, {first_non_bulk_layer.element})'
                f' and bottommost sublayer ({self.sublayers[-1].element})'
                )

        ori = first_non_bulk_layer.cartpos                                      # TODO: need to flip with .cartpos[2]? Issue #174
        repeat_vectors = []
        for atom in self.sublayers[-1]:
            test_v = atom.cartpos - ori          # From slab to bulk
            if self.is_translation_symmetric(test_v, eps, z_periodic=False,
                                             z_range=z_range):
                repeat_vectors.append(-test_v)   # From bulk to slab
        if not repeat_vectors:
            raise NoBulkRepeatError(
                f'{type(self).__name__}.identify_bulk_repeat: failed. '
                'No repeat vectors'
                )

        # Pick the shortest repeat vector, and return it with the
        # 'conventional' z direction (positive from bulk to surface)
        shortest = min(repeat_vectors, key=np.linalg.norm)
        shortest[2] *= -1                                                       # TODO: .cartpos[2] -- Issue #174
        return shortest

    def make_bulk_slab(self, rpars, recenter=True):
        """Return a copy of this slab with only bulk layers.

        Parameters
        ----------
        rpars : Rparams
            The PARAMETERS used for making the new slab. Attributes
            used (read-only): SUPERLATTICE, superlattice_defined,
            BULK_REPEAT, SYMMETRY_EPS.
        recenter : bool, optional
            Whether the fractional coordinates of the bulk slab
            that is created should be adjusted so that the topmost
            bulk atom is at the same position. This can safely be
            left to the default as long as the bulk repeat vector
            is known and correct. Default is True.

        Returns
        -------
        bulk_slab : Slab
            Copy of self with the correct bulk unit cell. The same
            slab can then be accessed as the `bulkslab` attribute
            of this slab. Notice that no attempt is made on making
            the returned slab have a 'minimal' unit cell in either
            in- or out-of-plane directions. If you want that, you
            can later call `self.ensure_minimal_bulk_ab_cell(rpars)`
            (makes `self.bulkslab` have smallest in-plane cell) and
            `self.bulkslab.ensure_minimal_c_vector(rpars)` (also
            makes the repeat vector shortest). If you already know
            the minimal bulk c vector and/or in-plane cell, you can
            also use `self.bulkslab.apply_bulk_cell_reduction(...)`.

        Raises
        ------
        MissingLayersError
            If this method is called on a slab that has no layers.
            This normally means that `create_layers` was not called
            before.
        TooFewLayersError
            If this method is called on a slab that has only one
            layer. Normally this means that there is a problem with
            the LAYER_CUTS.
        """
        if self.n_layers < 2:
            err_txt = ('Less than two layers detected. Check POSCAR '
                       'and consider modifying LAYER_CUTS.')
            _LOGGER.error(err_txt)
            exc = TooFewLayersError(err_txt)
            if not self.layers:
                exc = MissingLayersError('No layers detected. May need to '
                                         'create_layers or to full_update')
            raise exc

        # Construct bulk slab
        bulk_slab = BulkSlab.from_slab(self)
        bulk_slab.clear_symmetry_and_ucell_history()
        bulk_slab.atlist = AtomList(bulk_slab.bulk_atoms)
        bulk_slab.layers = bulk_slab.bulk_layers
        bulk_slab.sublayers = ()

        kwargs = {
            'eps': rpars.SYMMETRY_EPS,
            'epsz': rpars.SYMMETRY_EPS.z,
            'new_c_vec': self.get_bulk_repeat(rpars),
            'new_ab_cell': np.dot(np.linalg.inv(rpars.SUPERLATTICE),
                                  self.ab_cell.T),
            'recenter': recenter,
            'z_periodic': False,  # We cannot guarantee that it will be
            }
        bulk_slab.apply_bulk_cell_reduction(**kwargs)
        a_norm, b_norm = np.linalg.norm(bulk_slab.ab_cell.T, axis=1)
        if rpars.superlattice_defined and a_norm > b_norm + 1e-4:
            _LOGGER.warning(
                'The bulk unit cell defined by SUPERLATTICE does not '
                'follow standard convention: the first vector is longer '
                'than the second. Make sure beams are labelled correctly.'
                )
            rpars.setHaltingLevel(1)
        self.bulkslab = bulk_slab
        return bulk_slab

    def make_supercell(self, transform):
        """Return a copy of the slab replicated according to `transform`.

        The 'inverse' (i.e., leading to a size reduction) of this
        operation can be obtained by calling `make_subcell` with
        the same `transform`. Atoms in the duplicated replicas are
        marked as duplicates of those in this slab.

        Parameters
        ----------
        transform : numpy.ndarray
            Shape (2, 2). All elements must be (close to) integers.
            The transposed unit cell is transformed by multiplication
            with `transform` on the left. Notice that this is NOT the
            usual way of doing it, e.g., as VESTA does. The equivalent
            matrix in VESTA is `transform.T`.

        Returns
        -------
        super_slab : Slab
            Copy of this slab replicated according to `transform`.

        Raises
        ------
        NonIntegerMatrixError
            If `transform` has non-integer entries.
        SingularMatrixError
            If `transform` is singular.
        ValueError
            If `transform` has unacceptable shape.
        """
        super_slab = copy.deepcopy(self)
        try:
            repeats = leedbase.get_superlattice_repetitions(transform)
        except ValueError as exc:
            raise type(exc)(
                f'{type(self).__name__}.make_supercell: {exc}'
                ) from None

        if any(r > 1 for r in repeats):  # Need to duplicate atoms
            repeat_ranges = [range(r) for r in repeats]
            for atom in super_slab.atlist.copy():
                for i, j in itertools.product(*repeat_ranges):
                    # pylint: disable=compare-to-zero
                    if i == j == 0:  # Skip already existing atom
                        continue
                    # Duplicate the atom, and implicitly store it
                    # in super_slab. Then change its fractional pos
                    duplicate_atom = atom.duplicate()
                    duplicate_atom.pos[:2] += (i, j)

        # We now may have added new atoms that will have fractional
        # coordinates outside the original cell. Make sure we have
        # the correct Cartesian coordinates before transforming the
        # unit cell.
        super_slab.update_cartesian_from_fractional()
        super_slab.ab_cell[:] = super_slab.ab_cell.dot(transform.T)

        # Starting from the stored Cartesian coordinates, collapse
        # all atoms (incl. fractional coordinates) to the new cell
        super_slab.collapse_cartesian_coordinates()
        return super_slab

    def make_subcell(self, rpars, transform):
        """Return a subcell of this slab.

        This is the inverse of `make_supercell(transform)` with
        the same `transform` matrix.

        Parameters
        ----------
        rpars : Rparams
            The PARAMETERS used to determine how the copy is to be
            prepared. Attributes used: SYMMETRY_EPS.
            Both attributes are used as Cartesian tolerances for
            removal of duplicate atoms resulting from the reduction
            of size.
        transform : Sequence
            Shape (2, 2). The transformation matrix defining the
            subcell-to-supercell relationship. It should be an
            integer-valued matrix. The `transform` is inverted,
            then applied by left multiplication with the in-plane
            unit cell, when the unit cell has vectors on the rows.
            Default is None.

        Returns
        -------
        subcell_slab : Slab
            A new slab with reduced in-plane size according
            to `transform`. Duplicate atoms resulting from
            the size reduction are removed. Atoms that have
            been removed from the original slab as a result
            of the size reduction are marked as duplicates
            of those that remain in `subcell_slab`.

        Raises
        ------
        ValueError
            If `transform` is not a (2x2) matrix.
        ValueError
            If `transform` is an invalid transformation for
            this slab, that is, if slab is not a supercell
            with `transform` as its superlattice matrix.
        NonIntegerMatrixError
            If `transform` is not integer-valued.
        SlabError
            If reducing the current slab was unsuccessful.
            This is normally because either `slab.layers`
            or `rpars` were out of date.

        Notes
        -----
        This method is especially useful in case the slab
            is a supercell containing multiple copies of
            the same slab, i.e., if `get_minimal_ab_cell`
            gives a possible reduction. In that case,
            checking the symmetry on the smaller slab
            is more efficient.
        """
        subcell_slab = copy.deepcopy(self)
        subcell_slab.clear_symmetry_and_ucell_history()
        subcell_slab.update_cartesian_from_fractional()

        transform = ensure_integer_matrix(transform)
        if transform.shape != (2, 2):
            raise ValueError(f'Invalid transform.shape={transform.shape}. '
                             'Expected transform.shape=(2, 2)')
        if np.allclose(transform, np.identity(2)):
            return subcell_slab

        # Reduce dimensions in (x, y), then remove duplicates
        try:
            subcell_slab.ab_cell[:] = np.dot(self.ab_cell,
                                             np.linalg.inv(transform).T)
        except np.linalg.LinAlgError as exc:  # singular
            raise SingularMatrixError(f'transform={transform} '
                                      'is singular') from exc
        subcell_slab.collapse_cartesian_coordinates(update_origin=True)
        subcell_slab.remove_duplicate_atoms(rpars.SYMMETRY_EPS,
                                            rpars.SYMMETRY_EPS.z,
                                            other_slab=self)
        # Now make sure that transform actually gave a subcell
        supercell_atoms = subcell_slab.n_atoms * abs(np.linalg.det(transform))
        if self.n_atoms != supercell_atoms:
            raise ValueError(
                f'Slab is not a supercell with transform={transform} '
                'as its superlattice matrix'.replace('\n', ',')
                )
        return subcell_slab

    def restoreOriState(self, keepDisp=False):
        """Resets the atom positions and site vibrational amplitudes to the
        original state, and stores the deviations as offset instead."""
        for site in self.sitelist:
            if site.oriState is None:
                continue
            siteats = [at for at in self if at.site == site]
            for el in site.occ:
                for at in siteats:
                    o = site.vibamp[el] - site.oriState.vibamp[el]
                    if el in at.offset_vib:
                        o += at.offset_vib[el]
                    at.offset_vib[el] = o
                    if abs(at.offset_vib[el]) < 1e-6:
                        del at.offset_vib[el]
                site.vibamp[el] = site.oriState.vibamp[el]
                site.occ[el] = site.oriState.occ[el]
        uci = np.linalg.inv(self.ucell)
        self.update_cartesian_from_fractional()
        for at in self:
            if at.oriState is None:
                continue
            for el in at.disp_occ.keys():
                o = np.array([0., 0., 0.])
                if el in at.offset_geo:
                    o += at.offset_geo[el]
                if el in at.disp_geo_offset:
                    o += at.disp_geo_offset[el][0]
                else:
                    o += at.disp_geo_offset['all'][0]
                o[2] *= -1
                ro = np.dot(uci, o)
                nro = at.pos - at.oriState.pos + ro
                for i in range(0, len(nro)):  # minimize abs(nro)
                    while nro[i] > 0.5:
                        nro[i] -= 1
                    while nro[i] < -0.5:
                        nro[i] += 1
                no = np.dot(self.ucell, nro)
                no[2] *= -1
                at.offset_geo[el] = no
            at.pos = at.oriState.pos
            at.disp_geo_offset = {'all': [np.zeros(3)]}
        self.collapse_fractional_coordinates()
        self.update_cartesian_from_fractional()
        self.update_layer_coordinates()
        if keepDisp:
            return
        for at in self:
            at.known_deltas = []
            at.initDisp(force=True)
            at.constraints = {1: {}, 2: {}, 3: {}}
        return

    def with_extra_bulk_units(self, rpars, n_cells):
        """Return a copy with extra bulk unit cells added at the bottom.

        Parameters
        ----------
        rpars : Rparams
            The PARAMETERS  from which BULK_REPEAT is taken.
        n_cells : int
            The number of bulk unit cells to be added to the
            copy returned.

        Returns
        -------
        bulk_appended : Slab
            A copy of self with `n_cells` bulk cells added at the
            bottom.
        new_bulk_atoms : list
            The Atoms that were added to append the new bulk cells.

        Raises
        ------
        MissingLayersError
            If this method is called before any layer was identified,
            or if none of the layers is labelled as bulk.
        SlabError
            If the bulk repeat vector would give an unreasonably small
            interlayer distance. This may mean that (1) there are too
            few layers (e.g., only one thick layer), or (2) the bulk
            was not identified correctly (e.g., wrong BULK_REPEAT).
        ValueError
            If `n_cells` is not a positive number.

        Note
        ----
        Layers in the `bulk_appended` slab retuned are re-labelled
        so that there are exactly `rpars.N_BULK_LAYERS` bulk layers
        at the bottom. This means that layers of this slab that are
        labelled as bulk are turned into non-bulk layers in the
        `bulk_appended` slab returned (as are atoms in those layers).
        """
        if n_cells <= 0:
            raise ValueError('with_extra_bulk_units: n_cells must be positive')
        eps = 0.1
        bulk_appended = copy.deepcopy(self)
        bulk_layers = bulk_appended.bulk_layers
        if not bulk_layers:
            raise MissingLayersError('No bulk layers to duplicate')

        # First get the bulk repeat vector (with positive z).
        bulk_c = self.get_bulk_repeat(rpars)
        if abs(bulk_c[2]) < eps:
            raise SlabError('Bulk interlayer distance is too small: '
                            f'{bulk_c[2]}. Check LAYER_CUTS.')

        # Now take the component of bulk_c parallel to unit-cell c.
        # This will be used for expanding the unit cell.
        slab_c = bulk_appended.c_vector
        slab_c_direction = slab_c / np.linalg.norm(slab_c)
        bulk_c_par = bulk_c.dot(slab_c_direction) * slab_c_direction

        # Since we use the opposite-z convention for atoms, we also
        # need the two components with flipped z. We will move old
        # atoms up along unit-cell c, while new atoms will be shifted
        # only perpendicular to it.
        bulk_c[2] *= -1                                                         # TODO: .cartpos[2] -- Issue #174
        bulk_c_par_atoms = bulk_c.dot(slab_c_direction) * slab_c_direction
        bulk_c_perp_atoms = bulk_c - bulk_c_par_atoms

        bulk_appended.update_cartesian_from_fractional()
        new_bulk_atoms = []
        for _ in range(n_cells):
            # The trick to get it right is always adding atoms as
            # duplicates of the bottommost bulk layers, which may
            # either be the ones that this slab originally had, or
            # those that we have just added.
            # pylint: disable=protected-access
            (added_atoms,
             bulk_layers) = bulk_appended._add_one_bulk_cell(bulk_layers,
                                                             bulk_c_par,
                                                             bulk_c_par_atoms,
                                                             bulk_c_perp_atoms)
            new_bulk_atoms.extend(added_atoms)
        bulk_appended.collapse_cartesian_coordinates(update_origin=True)

        # Make sure that the BULK_REPEAT is not too short.
        # We can only cover the case in which the BULK_REPEAT
        # gives negative interlayer distances. See discussion
        # at https://github.com/viperleed/viperleed/issues/187
        if any(d < eps for d in bulk_appended.interlayer_gaps):
            raise SlabError('BULK_REPEAT is too small. Gives layers '
                            'with negative interlayer distances')

        # Make sure the number of bulk layers in the new slab is
        # still consistent with rpars.N_BULK_LAYERS. This means
        # turning all layers that were bulk before into 'non-bulk'
        for non_bulk_layer in bulk_appended.layers[:-rpars.N_BULK_LAYERS]:
            non_bulk_layer.is_bulk = False

        bulk_appended.sort_original()
        # Invalidate outdated information
        bulk_appended.sublayers = ()

        if not bulk_appended.is_bulk:
            bulk_appended.bulkslab = None
        return bulk_appended, new_bulk_atoms



"""Module beamgen of viperleed.calc.files

Creates the BEAMLIST file for TensErLEED calculations.

Original version by Florian Kraushofer (2020) was a wrapper for
the Fortran beamgen script by Uli Löffler et al. Complete refactor
by Alexander Imre (2023) that removes the dependency on the Fortran
script and instead accomplishes the same in python. This is more
flexible and allows for more accurate calculations.
"""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    'Florian Kraushofer (@fkraushofer)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-17'
__license__ = 'GPLv3+'

import logging

import fortranformat as ff
import numpy as np

from viperleed.calc import symmetry
from viperleed.calc.lib.leedbase import BOHR_TO_ANGSTROM
from viperleed.calc.lib.leedbase import HARTREE_TO_EV
from viperleed.calc.lib.leedbase import getLEEDdict
from viperleed.guilib.base import BeamIndex
from viperleed.guilib.base import get_equivalent_beams


H_BAR_SQ_OVER_2M = 0.5 * HARTREE_TO_EV * BOHR_TO_ANGSTROM**2  # h**2/2m
_LOGGER = logging.getLogger(__name__)


def calc_and_write_beamlist(slab, rpars, domains=False,
                            beamlist_name='BEAMLIST'):
    """Calculate and write the contents for file BEAMLIST.

    BEAMLIST contains a list of all diffraction beams that will be
    used for internal calculations (as opposed to IVBEAMS, which
    contains the beams to be output). The file lists the beams with
    indices as float values and a lower cutoff energy below which
    the beam is evanescent (i.e., does not leave the surface and
    has intensity 0). The used format is defined in make_beamlist_lines
    and dictated by TensErLEED and the legacy beamgen scripts that were
    used before. Note that BEAMLIST is not read directly by refcalc,
    but instead all input files for the refcalc are combined into one
    string (by collectFIN in iorefcalc.py) and then piped in.

    NB: The energies calculated here are slightly higher than those
    from beamgenv3 (and beamgen.v1.7) because we use *more accurate*
    values for unit conversions. The legacy code used these rounded
    values:
    HARTREE_TO_EV = 27.21
    BOHR_TO_ANGSTROM = 0.529
    Similarly, the list of included beams may be different for the same
    energy range, as the legacy code used rounded values and typecast a
    cutoff from float to int.

    In any case, this version should give more accurate energy values
    and be more generous in how many beams are considered.

    Parameters
    ----------
    slab : Slab
        Slab object. A .bulkslab will be added if it was not
        yet created for this slab.
    rpars : Rparams
        Run parameters.
    domains : bool, optional
        Flag to indicate if performing a domain calculation,
        by default False.
    beamlist_name : str or pathlike, optional
        Filename to be written, by default "BEAMLIST".
    """
    if slab.bulkslab is None:
        slab.make_bulk_slab(rpars)
        symmetry.findSymmetry(slab.bulkslab, rpars)

    # Use guilib to generate list of beams
    leed_parameters = getLEEDdict(slab, rpars)
    leed_parameters['screenAperture'] = 180  # All possible beams
    leed_parameters['eMax'] = _get_emax_for_evanescent_beams(slab, rpars,
                                                             domains)

    # use **only** beams from domain specified in rpars.SUPERLATTICE
    # beams come pre-sorted from get_equivalent_beams()
    equivalent_beams = get_equivalent_beams(leed_parameters, domains=0)

    # Log beam groups for debugging
    _log_beamgroups(equivalent_beams)

    # Strip away symmetry information, and sort into scattering subsets
    beam_subsets = _split_into_subsets([BeamIndex(hk)  # str->Fraction
                                        for hk, *_ in equivalent_beams])

    energies, indices = _prepare_beam_groups(slab.bulkslab,
                                             beam_subsets.values())

    # get highest energy considered; groups may have different shapes
    max_energy = max(np.max(group) for group in energies)
    _LOGGER.debug(f'Highest energy considered in '
                  f'BEAMLIST: {max_energy:.2f} eV')

    # write to file
    try:  # pylint: disable=too-many-try-statements
        with open(beamlist_name, 'w', encoding='utf-8') as file:
            file.writelines(
                make_beamlist_lines(indices, energies, rpars.TL_VERSION)
                )
    except OSError:  # Do not handle ValueError on purpose
        _LOGGER.error(f'Unable to write file {beamlist_name}')
        raise

    _LOGGER.debug('Wrote to BEAMLIST successfully.')


def _get_emax_for_evanescent_beams(slab, rpars, domains):
    """Return an energy cut-off that will generate also evanescent beams.

    The full-dynamic calculation considers scattering of propagating
    beams as well as evanescent ones. Among the latter, only those
    that survive attenuation when propagating between two layers are
    considered. Beams are considered surviving if their amplitude is
    attenuated by a factor less than ATTENUATION_EPS. Here we increase
    the highest energy by the corresponding factor.

    Parameters
    ----------
    slab : Slab
        The structure for which beams are to be calculated.
    rpars : Rparams
        Run parameters.
    domains : bool
        Flag to indicate if performing a domain calculation.

    Returns
    -------
    e_max : float
        The energy (in electronvolts) that generates also the
        correct evanescent beams.
    """
    if not domains:
        d_min = slab.smallest_interlayer_gap
    else:
        d_min = min(dp.sl.smallest_interlayer_gap for dp in rpars.domainParams)
    d_min *= 0.7                                                                # TODO: may want to complain if this is small as it will give a huge load of beams (and may mean different LAYER_CUTS should be used).

    e_max = rpars.THEO_ENERGIES.max
    e_max += H_BAR_SQ_OVER_2M * (np.log(rpars.ATTENUATION_EPS) / d_min)**2
    return e_max


def _log_beamgroups(equivalent_beams):
    """Log information on beam groups if desired."""
    if _LOGGER.getEffectiveLevel() > 5:
        return

    full_log_msg = ['Equivalent beams:',
                    '(   h     |   k     ),group,']
    for beam_hk, beam_group, *_ in equivalent_beams:
        index = BeamIndex(beam_hk)
        line = f'{format(index, "(4,4)s")}, {beam_group:4},'
        full_log_msg.append(line)

    # Split log message into two parts
    # The first 15 lines are intended for log-level verbose...
    _LOGGER.log(level=5, msg='\n'.join(full_log_msg[:15]))
    # ...the rest only at very verbose
    _LOGGER.log(level=1, msg='\n'+'\n'.join(full_log_msg[15:]))


def _prepare_beam_groups(bulk, beam_subsets):
    """Return energies and indices of beams in beam_subsets as groups.

    Parameters
    ----------
    bulk : BulkSlab
        The bulk slab to which the indices in beam_subsets refer.
    beam_subsets : Iterable
        One element per beam group. Each element should contain
        the indices of the beams to be prepared for the output.
        Each beam index is expected to be a 2-sequence of
        numbers, which will be converted to float.

    Returns
    -------
    energies : list
        len(energies) == len(beam_subsets). Each element is a
        numpy.ndarray of the vacuum energies (in electronvolts)
        of the beams in the group.
    indices : list
        len(indices) == len(energies). indices[i] is a numpy.ndarray
        version of the indices in beam_subsets[i], converted to float.
    """
    reciprocal_bulk = 2*np.pi*np.linalg.inv(bulk.ab_cell.T).T
    all_energies = []
    all_indices = []
    for beam_indices in beam_subsets:
        indices_arr = np.array(beam_indices, dtype='float64')  # from Fraction
        # Calculate cut-off energy for each beam
        energies = (                                                            # TODO: we could probably remove the energies from BEAMLIST completely. It seems they are not used in TensErLEED (see subroutine READIN in lib.tleed.f). Would need to remove it from here, and readBEAMLIST in beams.py.
            H_BAR_SQ_OVER_2M
            * np.sum(np.dot(indices_arr, reciprocal_bulk)**2, axis=1)
            )

        # generate file contents for beam subset
        all_indices.append(indices_arr)
        all_energies.append(energies)
    return all_energies, all_indices


def make_beamlist_lines(all_indices, all_energies, tl_version):
    """Yield lines to be written to file BEAMLIST.

    Beam sets that are not related by a bulk translation are
    separated into different "blocks". The format is the same
    as used be the legacy beamgen scripts by U. Loeffler and R.
    Doell. See also help(get_beam_scattering_subsets).

    Parameters
    ----------
    all_indices : list
        Elements are numpy arrays with shape (n_beams_subset, 2).
        Each element is one subset, containing the hk-indices
        (diffraction orders) of the beams. n_beams_subset is the
        number of beams in each subset.
    all_energies : list
        Lower cut-off energies for the beams. Each element is one
        subset. Elements are numpy arrays with shape (n_beams_subset,).
    tl_version : float
        Version of TensErLEED. To be taken from Rparams.TL_VERSION.
        This values decides the format of the output string.

    Yields
    ------
    beamlist_lines : str
        String representation of each line of the BEAMLIST file.
        '\n'-terminated.

    Raises
    ------
    ValueError
        If indices and energies have incompatible shapes.
    ValueError
        If one of the beam sets exceeds the maximum size
        currently supported in the Fortran code.
    """
    # Set up Fortran format as was used by beamgen.
    # TensErLEED v1.7 and higher used beamgen v1.7; earlier versions
    # used v3. This matters because the format changed slightly:
    # beamgen v1.7 had I5, beamgen v3 had I4. Also the maximum
    # number of beams per beamset has changed since.
    if tl_version < 1.7:
        fmt = {
            'beamset': ff.FortranRecordWriter(
                "2F10.5,2I3,10X,'E =  ',F10.4,2X,'NR.',I4"
                ),
            'n_beams': ff.FortranRecordWriter('15I4')
            }
        n_beams_max = 9999
    else:
        fmt = {
            'beamset': ff.FortranRecordWriter(
                "2F10.5,2I3,10X,'E =  ',F10.4,2X,'NR.',I5"
                ),
            'n_beams': ff.FortranRecordWriter('15I5')
            }
        n_beams_max = 99999

    if len(all_indices) != len(all_energies):
        raise ValueError(
            f'Inconsistent number of beam indices ({len(all_indices)}) '
            f'and energies ({len(all_energies)})'
            )
    beam_nr = 1
    for indices, energies in zip(all_indices, all_energies):
        n_beams = indices.shape[0]
        if energies.shape != (n_beams,) or indices.shape != (n_beams, 2):
            raise ValueError(
                f'Incompatible size of indices (shape={indices.shape})'
                f'and energies (shape={energies.shape}).'
                )

        # first line contains number of beams
        if n_beams > n_beams_max:
            raise ValueError(
                f'Too many beams in beam set ({n_beams}). TensErLEED v'
                f'{tl_version:.2f} code supports at most {n_beams_max}.'
                )
        yield fmt['n_beams'].write([n_beams]) + '\n'

        # iterate over all beams and format lines
        for beam_hk, energy in zip(indices, energies):
            line = fmt['beamset'].write([*beam_hk, 1, 1, energy, beam_nr])
            yield line + '\n'
            beam_nr += 1


def get_beam_scattering_subsets(beam_indices_raw):
    """Return beam scattering subsets and reduced indices for a list of beams.

    LEED diffraction beams are grouped into subsets for the computation
    of reflection/transmission matrices. In the full-dynamic scattering
    calculation (refcalc), one needs to consider that one beam can be
    scattered into another. However, this is only possible, if the
    beam wave-vectors are related by the *bulk* unit cell. I.e., beam
    (1/2, 0) can be scattered into (3/2, 0), but not into (1, 0).
    This property can be used in the refcalc to simplify calculations
    by making the reflection/transmission matrices block-diagonal.
    To enable this, we need to group the beams accordingly in BEAMLIST.

    This function takes a list of beams (as BeamIndex objects) and
    calculates reduced indices via h_red = h%1, k_red = k%1 (wrapping
    the beams back in the first Brillouin zone). It then takes a set
    of the reduced indices, to generate unique identifiers of the
    subsets. The first (and possibly only) subset contains, by
    definition, the integer beams, starting with (0|0).

    Parameters
    ----------
    beam_indices_raw : Sequence
        Beam indices in (h, k) form.

    Returns
    -------
    subset_classes : dict_keys
        The unique first-Brillouin-zone beam indices for the
        beam subsets, sorted in their order of appearance in
        `beam_indices_raw`. The length gives the number of
        subsets.
    reduced_indices : list
        Reduced version of the indices in beam_indices_raw.
    """
    reduced_indices = [(h % 1, k % 1) for (h, k) in beam_indices_raw]
    subset_classes = dict.fromkeys(reduced_indices)
    return subset_classes.keys(), reduced_indices


def _split_into_subsets(beam_indices_raw):
    """Split raw beam indices into bulk-independent subsets.

    See also help(get_beam_scattering_subsets).

    Parameters
    ----------
    beam_indices_raw : Sequence
        Beam indices in (h, k) form.

    Returns
    -------
    beam_subsets : dict
        Keys are the unique first-Brillouin-zone beam indices for
        the beam subsets. They are insertion-ordered by their norm.
        Values are all the indices in beam_indices_raw that belong
        to the specific subset.
    """
    (subset_classes,
     reduced_indices) = get_beam_scattering_subsets(beam_indices_raw)           # TODO: create test case to check that len(subset_classes) == np.linalg.det(rp.SUPERLATTICE)
    beam_subsets = {red_index: [] for red_index in subset_classes}
    for raw_index, red_index in zip(beam_indices_raw, reduced_indices):
        beam_subsets[red_index].append(raw_index)
    return beam_subsets
